import torch
import torch.distributions as tdist
import pyro
import pyro.distributions as dist
import pyro.distributions.constraints as constraints


def get_alpha(
    expected_guide_p, size_factor, sample_mask, a0, epsilon=1e-5, normalize_by_a0=True
):
    p = (
        expected_guide_p.permute(0, 2, 1) * size_factor[:, None, :]
    )  # (n_reps, n_guides, n_bins)
    if normalize_by_a0:
        a = (
            (p + epsilon / p.shape[-1])
            / (p.sum(axis=-1)[:, :, None] + epsilon)
            * a0[None, :, None]
        )
        a = (a * sample_mask[:, None, :]).clamp(min=epsilon)
        return a
    a = (p * sample_mask[:, None, :]).clamp(min=epsilon)
    return a


def get_std_normal_prob(
    upper_quantile: torch.Tensor,
    lower_quantile: torch.Tensor,
    mu: torch.Tensor,
    sd: torch.Tensor,
    mask=None,
) -> torch.Tensor:
    """
    Returns the probability that the normal distribution with mu and sd will
    lie between upper_quantile and lower_quantile of normal distribution
    centered at 0 and has scale sd.
    Arguments
    - mask: ignore index if 0 (=False).
    """
    inf_mask = upper_quantile == 1.0
    ninf_mask = lower_quantile == 0.0

    upper_thres = torch.ones_like(upper_quantile)
    lower_thres = torch.zeros_like(lower_quantile)
    upper_thres[~inf_mask] = tdist.Normal(0, 1).icdf(upper_quantile[~inf_mask])
    lower_thres[~ninf_mask] = tdist.Normal(0, 1).icdf(lower_quantile[~ninf_mask])

    if not mask is None:
        # Temporarily add mask value
        sd = sd + (~mask).long() * 100
    assert (sd > 0).all(), f"sd > 0 not met at {sd[~(sd > 0)]}"
    cdf_upper = torch.ones_like(upper_quantile)
    cdf_lower = torch.zeros_like(lower_quantile)
    assert (
        cdf_upper.shape == inf_mask.shape
    ), f"mask:{inf_mask.shape}, cdf:{cdf_upper.shape}"

    cdf_upper[~inf_mask] = tdist.Normal(mu[~inf_mask], sd[~inf_mask]).cdf(
        upper_thres[~inf_mask]
    )
    cdf_lower[~ninf_mask] = tdist.Normal(mu[~ninf_mask], sd[~ninf_mask]).cdf(
        lower_thres[~ninf_mask]
    )
    res = cdf_upper - cdf_lower
    if not mask is None:
        res[~mask] = 0

    return res


def _scale_edited_pi(
    pi: torch.Tensor,
    guide_accessibility: torch.Tensor,
    a: float = 0.2513,
    b: float = -1.9458,
):
    """Scale editied pi by its accessibility.

    Data fitted through relationship observed from data (updated 1/17/2023).
    Transformation derived from linear regression of log(endo/reporter) ~ a*log(atac_signal + 1)+b.
    If pi of multiple alleles are provided, pi is scaled so that total scaled pi wouldn't exceed 1.0.

    Args
    pi: Editing rate
    guide_accessibility: raw accessibility score
    """
    assert pi.shape[-2] == guide_accessibility.shape[-1], (
        pi.shape,
        guide_accessibility.shape,
    )
    return (
        pi
        * torch.exp(torch.tensor(b))
        * torch.pow(guide_accessibility, a).unsqueeze(-1)
    )


def scale_pi_by_accessibility(
    pi: torch.Tensor,
    guide_accessibility: torch.Tensor,
    a: float = 0.2513,
    b: float = -1.9458,
    fit_noise: bool = False,
):
    """Scale pi by its accessibility.

    Data fitted through relationship observed from data (updated 1/17/2023).
    Transformation derived from linear regression of log(endo/reporter) ~ a*log(atac_signal + 1)+b.
    If pi of multiple alleles are provided, pi is scaled so that total scaled pi wouldn't exceed 1.0.

    Args
    pi: Editing rate (Has control editing rate at 0th index of last dimension)
    guide_accessibility: raw accessibility score for each guide. Shape (n_guides,)
    """
    pi_shape = pi.shape
    scaled_pi = _scale_edited_pi(pi[..., 1:], guide_accessibility)
    ctrl_pi = torch.ones(pi[:, :, :, 0].shape) - scaled_pi.sum(axis=-1)
    pi = torch.concat([ctrl_pi.unsqueeze(-1), scaled_pi], axis=-1)
    pi = pi / pi.sum(axis=-1).clamp(min=1.0)[..., None]
    assert pi.shape == pi_shape, (pi.shape, scaled_pi.shape, pi_shape)
    pi = add_noise_to_pi(pi, fit_noise=fit_noise)
    return pi


def add_noise_to_pi(pi: torch.Tensor, pi_noise_sd: float = 0.655, fit_noise=False):
    """Add noise to pi.

    Gaussian noise is added on the logit scale. Scale of the noise is fitted from the data (updated 1/17/2023).

    Args
    pi: Editing rate of shape (n_reps, 1, n_guides, n_alleles). Has control editing rate at 0th index of last dimension.
    """
    pi_shape = pi.shape
    n_reps, _, n_guides, n_alleles = pi_shape
    logit_pi = torch.logit(pi[..., 1:].clamp(min=1e-3, max=1 - 1e-3))
    if fit_noise:
        with pyro.plate("guide_plate_noise", n_guides):
            noise_loc = pyro.param("noise_loc", torch.zeros((n_guides,)))
            noise_scale = pyro.param(
                "noise_scale",
                torch.ones((n_guides,)) * pi_noise_sd,
                constraint=constraints.positive,
            )
            logit_pi_noise = pyro.sample(
                "logit_pi_noise",
                dist.Normal(noise_loc, noise_scale),  # infer={"is_auxiliary": True}
            )
    else:
        with pyro.plate("guide_plate_noise", n_guides):
            logit_pi_noise = pyro.sample(
                "logit_pi_noise",
                dist.Normal(0, pi_noise_sd),  # infer={"is_auxiliary": True}
            )
    logit_pi += (
        logit_pi_noise.unsqueeze(0)
        .unsqueeze(0)
        .unsqueeze(-1)
        .expand((n_reps, 1, -1, n_alleles - 1))
    )
    exp_pi_noised = torch.exp(logit_pi)
    pi_noised = (exp_pi_noised / (1 + exp_pi_noised)).clamp(min=1e-3, max=1 - 1e-3)
    pi = torch.concat(
        [
            (torch.ones(pi[:, :, :, 0].shape) - pi_noised.sum(axis=-1)).unsqueeze(-1),
            pi_noised,
        ],
        axis=-1,
    )
    assert pi.shape == pi_shape, pi.shape
    return pi



from copy import deepcopy
from typing import List, Tuple
from tqdm.auto import tqdm
from ..framework.Edit import Edit, Allele
from ..framework.AminoAcidEdit import CodingNoncodingAllele
import pandas as pd
from ..annotate.translate_allele import CDS, RefBaseMismatchException


def filter_allele_by_pos(
    allele: Allele,
    pos_start: int = None,
    pos_end: int = None,
    filter_rel_pos=True,
):
    """
    Filter alleles based on position and return the filtered allele and
    number of filtered edits.
    --
    Keyword arguments
    pos_start (int) -- start position to include (inclusive)
    pos_end (int) -- end position to include (exclusive)
    """
    filtered_edits = 0
    allele_filtered = deepcopy(allele)
    if not (pos_start is None and pos_end is None):
        if filter_rel_pos:
            for edit in allele.edits:
                if not (edit.rel_pos >= pos_start and edit.rel_pos < pos_end):
                    filtered_edits += 1
                    allele_filtered.edits.remove(edit)
        else:
            for edit in allele.edits:
                if not (edit.pos >= pos_start and edit.pos < pos_end):
                    filtered_edits += 1
                    allele_filtered.edits.remove(edit)

    else:
        print("No threshold specified")  # TODO: warn
    return (allele_filtered, filtered_edits)


def filter_allele_by_base(
    allele: Allele,
    allowed_base_changes: List[Tuple] = None,
    allowed_ref_base: str = None,
    allowed_alt_base: str = None,
):
    """
    Filter alleles based on position and return the filtered allele and
    number of filtered edits.
    """
    filtered_edits = 0
    if (
        not (allowed_ref_base is None and allowed_alt_base is None)
        + (allowed_base_changes is None)
        == 1
    ):
        print("No filters specified or misspecified filters.")
    elif not allowed_base_changes is None:
        for edit in allele.edits.copy():
            if not (edit.ref_base, edit.alt_base) in allowed_base_changes:
                filtered_edits += 1
                allele.edits.remove(edit)
    elif not allowed_ref_base is None:
        for edit in allele.edits.copy():
            if edit.ref_base != allowed_ref_base:
                filtered_edits += 1
                allele.edits.remove(edit)
            elif not allowed_alt_base is None and edit.alt_base != allowed_alt_base:
                filtered_edits += 1
                allele.edits.remove(edit)
    else:
        for edit in allele.edits.copy():
            if edit.alt_base != allowed_alt_base:
                filtered_edits += 1
                allele.edits.remove(edit)
    return (allele, filtered_edits)


def get_aa_alleles(allele_str, include_synonymous=True):
    ldlr_cds = CDS()
    try:
        ldlr_cds.edit_allele(allele_str)
        ldlr_cds.get_aa_change(include_synonymous)
    except RefBaseMismatchException as e:
        print(e)
        return "ref mismatch"
    return ldlr_cds.get_aa_change(True)


def map_alleles_to_filtered(
    raw_allele_counts: pd.DataFrame,
    filtered_allele_counts: pd.DataFrame,
    jaccard_threshold=0.5,
):
    mapped_allele_counts = []  # pd.DataFrame(columns=raw_allele_counts.columns)
    for guide, guide_raw_counts in tqdm(
        raw_allele_counts.groupby("guide"),
        desc="Mapping alleles to closest filtered alleles",
    ):
        guide_filtered_allele_counts = filtered_allele_counts.loc[
            filtered_allele_counts.guide == guide, :
        ].set_index("allele")
        guide_filtered_alleles = guide_filtered_allele_counts.index.tolist()
        if len(guide_filtered_alleles) == 0:
            pass
        else:
            guide_raw_counts["allele_mapped"] = guide_raw_counts.allele.map(
                lambda allele: allele.map_to_closest(
                    guide_filtered_alleles,
                    jaccard_threshold=jaccard_threshold,
                    merge_priority=guide_filtered_allele_counts.mean(
                        axis=1, numeric_only=True
                    ),
                )
            )
            # prioritize allele with most mean counts
            guide_raw_counts = guide_raw_counts.drop("allele", axis=1).rename(
                columns={"allele_mapped": "allele"}
            )
            guide_raw_counts = guide_raw_counts.groupby(["guide", "allele"]).sum()
            mapped_allele_counts.append(guide_raw_counts)
    res = pd.concat(mapped_allele_counts).reset_index()
    res = res.loc[res.allele.map(str) != "", :]
    return res


def _map_alleles_to_filtered(
    raw_allele_counts: pd.DataFrame,
    filtered_allele_counts: pd.DataFrame,
    jaccard_threshold=0.5,
    aa_jaccard_threshold=0.5,
    nt_jaccard_threshold=0.5,
    allele_col="allele",
):
    """
    Map pre-filtering alleles to the closest post-filtering alleles and aggregate the read count.
    """
    mapped_allele_counts = []
    is_cn_allele = isinstance(
        raw_allele_counts.reset_index()[allele_col][0], CodingNoncodingAllele
    )
    for guide, guide_raw_counts in tqdm(
        raw_allele_counts.groupby("guide"),
        desc="Mapping alleles to closest filtered alleles",
    ):

        guide_filtered_allele_counts = filtered_allele_counts.loc[
            filtered_allele_counts.guide == guide, :
        ].set_index(allele_col)
        guide_filtered_alleles = guide_filtered_allele_counts.index.tolist()
        if len(guide_filtered_alleles) == 0:
            pass
        else:
            # prioritize allele with most mean counts
            merge_priority = guide_filtered_allele_counts.mean(
                axis=1, numeric_only=True
            )
            if is_cn_allele:
                guide_raw_counts["allele_mapped"] = guide_raw_counts[allele_col].map(
                    lambda allele: allele.map_to_closest(
                        guide_filtered_alleles,
                        aa_jaccard_threshold=aa_jaccard_threshold,
                        nt_jaccard_threshold=nt_jaccard_threshold,
                        merge_priority=merge_priority,
                    )
                )
            else:
                guide_raw_counts["allele_mapped"] = guide_raw_counts[allele_col].map(
                    lambda allele: allele.map_to_closest(
                        guide_filtered_alleles,
                        jaccard_threshold=jaccard_threshold,
                        merge_priority=merge_priority,
                    )
                )
            guide_raw_counts = (
                guide_raw_counts.drop(allele_col, axis=1)
                .rename(columns={"allele_mapped": allele_col})
                .groupby(["guide", allele_col])
                .sum()
            )

            mapped_allele_counts.append(guide_raw_counts)
    res = pd.concat(mapped_allele_counts).reset_index()
    res = res.loc[res[allele_col].map(bool), :]
    return res



"""Plotting functions to describe allele/guide/edit stats for allele count information stored in ReporterScreen.uns[allele_df_key]."""
import numpy as np
import matplotlib.pyplot as plt


def plot_n_alleles_per_guide(
    bdata, allele_df_key: str, allele_col: str = "allele", ax=None
):
    guide_to_allele = dict(
        list(
            bdata.uns[allele_df_key][["guide", allele_col]].groupby("guide")[allele_col]
        )
    )
    lens = [
        len(guide_to_allele[g]) if g in guide_to_allele else 0
        for g in bdata.guides.index
    ]
    ax.hist(lens, bins=np.arange(min(lens) - 0.5, max(lens) + 0.5))
    ax.set_xlabel("# alleles")
    ax.set_title(f"# Alleles per guide, raw (n={len(bdata.uns[allele_df_key])})")
    ax.set_ylabel("# guides")
    return ax


def plot_n_guides_per_edit(
    bdata, allele_df_key: str, allele_col: str = "aa_allele", ax=None
):
    a = bdata.uns[allele_df_key][["guide", allele_col]].copy()
    if allele_col == "aa_allele":
        a["edits"] = a[allele_col].map(
            lambda a: list(a.aa_allele.edits)
            + list(map(lambda e: e.get_abs_edit(), a.nt_allele.edits))
        )
    elif allele_col == "allele":
        a["edits"] = a[allele_col].map(lambda a: list(a.edits))
    edits_df = a.explode("edits")[["guide", "edits"]]
    edits_df["edits"] = edits_df.edits.map(str)
    edits_df = edits_df.loc[~edits_df.edits.str.startswith("-"), :].drop_duplicates()
    n_guides = edits_df.groupby("edits")["guide"].count()
    if ax is None:
        fig, ax = plt.subplots()
    ax.hist(n_guides, bins=np.arange(min(n_guides) - 0.5, max(n_guides) + 0.5))
    ax.set_xlabel("# guides")
    ax.set_title(f"# guides per edit (n={len(edits_df)})")
    ax.set_ylabel("# edits")
    return ax



import sys
import abc
from dataclasses import dataclass
from typing import Dict, Tuple
import logging
from xmlrpc.client import Boolean
from copy import deepcopy
import torch
import numpy as np
import pandas as pd
import bean as be
from .get_alpha0 import get_fitted_alpha0, get_pred_alpha0
from .get_pi_alpha0 import get_fitted_alpha0 as get_fitted_pi_alpha0
from .get_pi_alpha0 import get_pred_alpha0 as get_pred_pi_alpha0
from .utils import (
    Alias,
    get_accessibility_guides,
    get_edit_to_index_dict,
    _assign_rep_ids_and_sort,
)

logging.basicConfig(
    level=logging.INFO,
    format="%(levelname)-5s @ %(asctime)s:\n\t %(message)s \n",
    datefmt="%a, %d %b %Y %H:%M:%S",
    stream=sys.stderr,
    filemode="w",
)
error = logging.critical
warn = logging.warning
debug = logging.debug
info = logging.info


@dataclass
class ScreenData(abc.ABC):
    def __init__(
        self,
        screen: be.ReporterScreen,
        repguide_mask: str = None,
        sample_mask_column: str = None,
        shrink_alpha: bool = False,
        condition_column: str = "sort",
        control_condition: str = "bulk",
        accessibility_col: str = None,
        accessibility_bw_path: str = None,
        device: str = None,
        replicate_column: str = "rep",
        pi_popt: Tuple[float] = None,
        **kwargs,
    ):
        # TODO: remove replicate with too small number of (ex. only 1) sorting bin
        self.condition_column = condition_column
        self.device = device
        screen.samples["size_factor"] = self.get_size_factor(screen.X)
        if not (
            "rep" in screen.samples.columns
            and condition_column in screen.samples.columns
        ):
            screen.samples["rep"], screen.samples[condition_column] = zip(
                *screen.samples.index.map(lambda s: s.rsplit("_", 1))
            )
        if condition_column not in screen.samples.columns:
            screen.samples[condition_column] = screen.samples["index"].map(
                lambda s: s.split("_")[-1]
            )

        self.screen = screen
        self.screen_selected = screen[
            :, screen.samples[condition_column] != control_condition
        ]

        self.n_condits = len(
            self.screen_selected.var[condition_column].unique()
        )  # excluding bulk
        self.screen_control = screen[
            :, screen.samples[condition_column] == control_condition
        ]
        self.n_samples = len(screen.samples)  # 8
        self.n_guides = len(screen.guides)
        self.n_reps = len(screen.samples[replicate_column].unique())
        self.accessibility_col = accessibility_col
        self.accessibility_bw_path = accessibility_bw_path
        self.replicate_column = replicate_column
        self.condition_column = condition_column
        self.control_condition = control_condition
        self.sample_mask_column = sample_mask_column
        self.repguide_mask = repguide_mask
        self.shrink_alpha = shrink_alpha

    def _post_init(
        self,
    ):
        # Assign accessibility info
        if self.accessibility_col is not None:
            self.guide_accessibility = torch.as_tensor(
                self.screen.guides[self.accessibility_col].values
            )
        elif self.accessibility_bw_path is not None:
            self.guide_accessibility = get_accessibility_guides(
                self.accessibility_bw_path, self.screen.guides
            )
        else:
            self.guide_accessibility = None

        # Mask counts
        if self.sample_mask_column is not None:
            self.sample_mask = torch.as_tensor(
                self.screen_selected.samples[self.sample_mask_column].to_numpy()
            ).reshape(self.n_reps, self.n_condits)
            self.bulk_sample_mask = torch.as_tensor(
                self.screen_control.samples[self.sample_mask_column].to_numpy()
            )
        else:
            self.sample_mask = torch.ones((self.n_reps, self.n_condits), dtype=Boolean)
            self.bulk_sample_mask = torch.ones(self.n_reps, dtype=Boolean)
        self.X = self.transform_data(
            self.screen_selected.X
        )  # (n_reps, n_bins, n_guides)
        self.X_masked = self.X * self.sample_mask[:, :, None]
        self.X_control = self.transform_data(self.screen_control.X, 1)
        self.X_control_masked = self.X_control * self.bulk_sample_mask[:, None, None]
        if self.repguide_mask is None:
            self.repguide_mask = ~(self.X == 0).any(axis=1)
        else:
            print(f"Using repguide mask {self.repguide_mask}")
            assert (
                self.repguide_mask in self.screen.uns.keys()
            ), f"{self.repguide_mask} not in screen.uns"
            assert self.screen_selected.uns[self.repguide_mask].shape == (
                self.n_guides,
                self.n_reps,
            )
            assert (
                self.screen_selected.uns[self.repguide_mask].index
                == self.screen_selected.guides.index
            ).all()
            assert (
                self.screen_selected.uns[self.repguide_mask].columns
                == self.screen_selected.samples.rep.unique()
            ).all()
            self.repguide_mask = (
                torch.as_tensor(self.screen_selected.uns[self.repguide_mask].values.T)
                > 0
            )
            self.repguide_mask = torch.logical_and(
                self.repguide_mask, ~(self.X == 0).any(axis=1)
            )

        self.size_factor = torch.as_tensor(
            self.screen_selected.samples["size_factor"].to_numpy()
        ).reshape(self.n_reps, self.n_condits)
        self.size_factor_control = torch.as_tensor(
            self.screen_control.samples["size_factor"].to_numpy()
        ).reshape(self.n_reps, 1)
        # Get a0
        fitted_a0, self.popt = get_fitted_alpha0(
            self.X.clone().cpu(),
            self.size_factor.clone().cpu(),
            self.sample_mask.cpu(),
            shrink=self.shrink_alpha,
        )
        fitted_a0 = torch.as_tensor(fitted_a0)
        a0 = fitted_a0
        self.a0 = torch.as_tensor(a0)

    def __getitem__(self, guide_idx):
        ndata = deepcopy(self)
        ndata.screen_selected = ndata.screen_selected[guide_idx, :]
        ndata.screen_control = ndata.screen_control[guide_idx, :]
        ndata.n_guides = len(guide_idx)
        ndata.X = ndata.X[:, :, guide_idx]  # (n_reps, n_bins, n_guides)
        ndata.X_masked = ndata.X_masked[:, :, guide_idx]
        ndata.X_control = ndata.X_control[:, :, guide_idx]
        ndata.repguide_mask = ndata.repguide_mask[:, guide_idx]
        return ndata

    def transform_data(self, X, n_bins=None):
        if n_bins is None:
            n_bins = self.n_condits
        x = torch.as_tensor(X).T.reshape((self.n_reps, n_bins, self.n_guides)).float()
        if self.device is not None:
            x = x.cuda()
        return x

    def get_size_factor(self, X: np.array):
        """
        Get size factor for samples.
        """
        n_guides, n_samples = X.shape
        geom_mean_x = np.exp((1 / n_samples) * np.log(X + 0.5).sum(axis=1))
        assert geom_mean_x.shape == (n_guides,)
        norm_count = X / geom_mean_x[:, None]
        size_factor = np.median(norm_count, axis=0)
        assert size_factor.shape == (n_samples,)
        return size_factor

    def get_sample_ids_and_sort(self, screen: be.ReporterScreen, condit_id_col: str):
        screen.samples["rep_id"] = -1
        for i, rep in enumerate(sorted(screen.samples.rep.unique())):
            screen.var.loc[screen.samples.rep == rep, "rep_id"] = i
        if condit_id_col in screen.samples.columns:
            screen = screen[
                :, screen.samples.sort_values(["rep_id", condit_id_col]).index
            ]
        else:
            raise ValueError("AnnData doesn't have condition id provided.")
        return screen


@dataclass
class ReporterScreenData(ScreenData):
    X_bcmatch: torch.Tensor
    size_factor_bcmatch: torch.Tensor
    X_control_bcmatch: torch.Tensor
    size_factor_control_bcmatch: torch.Tensor
    allele_counts: torch.Tensor
    allele_counts_control: torch.Tensor
    a0_bcmatch: torch.Tensor
    a0_allele: torch.Tensor
    pi_a0: torch.Tensor

    def __init__(
        self,
        screen: be.ReporterScreen,
        repguide_mask: str = None,
        sample_mask_column: str = None,
        pi_popt: Tuple[float] = None,
        impute_pi_popt: bool = False,
        shrink_alpha: bool = False,
        condition_column: str = "sort",
        control_condition: str = "bulk",
        accessibility_bw_path: str = None,
        accessibility_col: str = None,
        use_const_pi: bool = False,
        pi_prior_count: int = 10,
        **kwargs,
    ):
        ScreenData.__init__(
            self,
            screen=screen,
            repguide_mask=repguide_mask,
            sample_mask_column=sample_mask_column,
            shrink_alpha=shrink_alpha,
            condition_column=condition_column,
            control_condition=control_condition,
            pi_popt=pi_popt,
            **kwargs,
        )
        self._post_init(
            screen,
            use_const_pi,
            impute_pi_popt,
            pi_prior_count,
            shrink_alpha,
            pi_popt,
        )

    def _post_init(
        self,
        screen,
        use_const_pi,
        impute_pi_popt,
        pi_prior_count,
        shrink_alpha,
        pi_popt,
    ):
        screen.samples["size_factor_bcmatch"] = self.get_size_factor(
            screen.layers["X_bcmatch"]
        )
        self.screen_selected.samples["size_factor_bcmatch"] = screen.samples.loc[
            self.screen_selected.samples.index, "size_factor_bcmatch"
        ]
        self.screen_control.samples["size_factor_bcmatch"] = screen.samples.loc[
            self.screen_control.samples.index, "size_factor_bcmatch"
        ]
        self.X_bcmatch = self.transform_data(self.screen_selected.layers["X_bcmatch"])
        self.X_bcmatch_masked = self.X_bcmatch * self.sample_mask[:, :, None]
        self.X_bcmatch_control = self.transform_data(
            self.screen_control.layers["X_bcmatch"], 1
        )
        self.X_bcmatch_control_masked = (
            self.X_bcmatch_control * self.bulk_sample_mask[:, None, None]
        )
        self.size_factor_bcmatch = torch.as_tensor(
            self.screen_selected.samples["size_factor_bcmatch"].to_numpy()
        ).reshape(self.n_reps, self.n_condits)
        self.size_factor_bcmatch_control = torch.as_tensor(
            self.screen_control.samples["size_factor_bcmatch"].to_numpy()
        ).reshape(self.n_reps, 1)

        edited_control = self.transform_data(
            self.screen_control.layers["edits"], n_bins=1
        )
        nonedited_control = self.X_bcmatch_control - edited_control
        nonedited_control[nonedited_control < 0] = 0
        if (
            not hasattr(self, "allele_counts_control")
            or self.allele_counts_control is None
        ):
            self.allele_counts_control = torch.stack(
                [nonedited_control, edited_control], axis=-1
            )  # (n_reps, n_bins, n_guides, n_alleles)
        # assert (self.allele_counts_control.sum(axis=-1) == self.X_bcmatch_control).all()
        a0_bcmatch = get_pred_alpha0(
            self.X_bcmatch.clone().cpu(),
            self.size_factor_bcmatch.clone().cpu(),
            self.popt,
            self.sample_mask.cpu(),
        )
        self.a0_bcmatch = torch.as_tensor(a0_bcmatch)

        if use_const_pi:
            self.pi = (self.allele_counts.sum(axis=(0, 1)) + pi_prior_count * 0.5) / (
                self.allele_counts.sum(axis=(0, 1)).sum(axis=-1)[..., None]
                + pi_prior_count
            )  # (n_guides, n_alleles)
            self.allele_counts_control = (
                self.X_bcmatch_control[:, :, :, None] * self.pi[None, None, :, :]
            )
            impute_pi_popt = True
        if impute_pi_popt:
            pi_popt = self.popt
        self.set_fitted_pi_a0(pi_popt, shrink_alpha, use_fitted_value=True)

    def __getitem__(self, guide_idx):
        ndata = super().__getitem__(guide_idx)
        ndata.X_bcmatch = ndata.X_bcmatch[:, :, guide_idx]
        ndata.X_bcmatch_masked = ndata.X_bcmatch_masked[:, :, guide_idx]
        ndata.X_bcmatch_control = ndata.X_bcmatch_control[:, :, guide_idx]
        if hasattr(ndata, "allele_counts"):
            ndata.allele_counts = ndata.allele_counts[:, :, guide_idx, :]
        if hasattr(ndata, "a0_allele"):
            ndata.a0_allele = ndata.a0_allele[guide_idx, :]
        if hasattr(ndata, "a0") and self.a0 is not None:
            ndata.a0 = ndata.a0[guide_idx]
        if hasattr(ndata, "pi_a0") and self.pi_a0 is not None:
            ndata.pi_a0 = ndata.pi_a0[guide_idx]
        if hasattr(ndata, "a0_bcmatch") and self.a0_bcmatch is not None:
            ndata.a0_bcmatch = ndata.a0_bcmatch[guide_idx]
        if hasattr(ndata, "pi_a0_bcmatch") and self.pi_a0_bcmatch is not None:
            ndata.pi_a0_bcmatch = ndata.pi_a0_bcmatch[guide_idx]
        return ndata

    def set_fitted_pi_a0(
        self, pi_popt: Tuple, shrink_alpha: bool = False, use_fitted_value: bool = True
    ):
        """Set pi_a0 values

        Args
        pi_popt: pre-fitted values
        shrink_alpha: Return shrinked alpha based on fitted trend, instead of fitted value itself.
        use_fitted_value: use fitted value. If False, use raw and nan values are substituted to fitted values.
        """
        if pi_popt is not None:
            pi_a0 = get_pred_pi_alpha0(
                self.allele_counts_control.clone().cpu(),
                self.size_factor_control.clone().cpu(),
                pi_popt,
            )
        else:
            pi_a0, _pi_popt = get_fitted_pi_alpha0(
                self.allele_counts_control.clone().cpu(),
                self.size_factor_control.clone().cpu(),
                fit_quantile=None,
                shrink=shrink_alpha,
            )
            self._pi_popt = _pi_popt

        if not use_fitted_value:
            pi_a0 = torch.as_tensor(
                get_fitted_pi_alpha0(
                    self.allele_counts_control.clone().cpu(),
                    self.size_factor_control.clone().cpu(),
                    fit=False,
                    fit_quantile=None,
                    shrink=shrink_alpha,
                )[0]
            )
            pi_a0[torch.isnan(pi_a0)] = self.pi_a0[torch.isnan(pi_a0)]
            self.pi_a0 = pi_a0
        self.pi_a0 = torch.as_tensor(pi_a0)


@dataclass
class VariantScreenData(ScreenData):
    n_targets: int
    n_sgRNAs_per_target: int
    target_lengths: torch.Tensor

    def __init__(
        self,
        screen: be.ReporterScreen,
        repguide_mask: str = None,
        sample_mask_column: str = None,
        shrink_alpha: bool = False,
        condition_column: str = "sort",
        control_condition: str = "bulk",
        target_col: str = "target",
        **kwargs,
    ):
        max_n_guides_per_target = screen.guides.groupby(target_col).size().max()
        info(
            f"Data has max {max_n_guides_per_target} guides per target element in '{target_col}' column."
        )

        ScreenData.__init__(
            self,
            screen=screen,
            repguide_mask=repguide_mask,
            sample_mask_column=sample_mask_column,
            shrink_alpha=shrink_alpha,
            condition_column=condition_column,
            control_condition=control_condition,
            **kwargs,
        )
        self.n_targets = len(screen.guides[target_col].unique())
        self.n_sgRNAs_per_target = max_n_guides_per_target
        self.target_lengths = self.get_target_lengths(screen, target_col)
        self.target_col = target_col

    def _post_init(self, target_col):
        self.target_col = target_col
        max_n_guides_per_target = self.screen.guides.groupby(target_col).size().max()
        info(
            f"Data has max {max_n_guides_per_target} guides per target element in '{target_col}' column."
        )
        self.n_targets = len(self.screen.guides[target_col].unique())
        self.n_sgRNAs_per_target = max_n_guides_per_target
        self.target_lengths = self.get_target_lengths(self.screen, self.target_col)

    def __getitem__(self, guide_idx):
        ndata = super().__getitem__(guide_idx)
        ndata.n_targets = len(ndata.screen_selected.guides[self.target_col].unique())
        ndata.target_lengths = ndata.get_target_lengths(
            ndata.screen_selected, self.target_col
        )
        return ndata

    def get_target_lengths(self, screen, target_col="target"):
        target_len_list = []
        screen.guides[target_col] = screen.guides[target_col].astype("category")
        cur_item = screen.guides[target_col].cat.codes[0]
        cur_len = 0
        n_targets = 0
        for i in screen.guides[target_col].cat.codes:
            if cur_item == i:
                cur_len += 1
            else:
                n_targets += 1
                target_len_list.append(cur_len)
                cur_len = 1
                cur_item = i
        n_targets += 1
        if n_targets != len(screen.guides[target_col].unique()):
            raise ValueError(
                f"Input Screen object not sorted for target identity. Sort the screen object so that guides targeting the same object would occur as consecutive block by screen[screen.guides[{target_col}].argsort(),:]"
            )  # TODO: sort within the script
        target_len_list.append(cur_len)
        target_lengths = torch.tensor(target_len_list)
        return target_lengths


@dataclass
class TilingReporterScreenData(ReporterScreenData):
    def __init__(
        self,
        screen: be.ReporterScreen,
        repguide_mask: str = None,
        sample_mask_column: str = None,
        shrink_alpha: bool = False,
        condition_column: str = "sort",
        control_condition: str = "bulk",
        allele_df_key="allele_counts_filtered",
        edit_index=None,
        alleles_to_select: pd.MultiIndex = None,
        allele_col: str = None,
        **kwargs,
    ):
        """
        Args
            screen: An object of type be.ReporterScreen, representing the screen data.
            repguide_mask: A string specifying the mask for selecting guides x reps DataFrame in screen.uns[repguide_mask].
            sample_mask_column: A string specifying the mask column in screen.samples[sample_mask_column].
            shrink_alpha: A boolean flag that shrinks alpha towards fitted line if True (dev).
            condition_column: Column name in screen.samples that specifies condition that is selected on.
            control_condition: A string specifying the control condition in screen.samples[condition_column].
            allele_df_key: A string representing the key for accessing the allele count dataframe in screen.uns.
            edit_index: A Pandas DataFrame that specifies the edit sequences to index.
            alleles_to_select: A Pandas multi-index representing the alleles that need to be subsetted.
            allele_col: A string representing the name of the allele column in the allele count dataframe (screen.uns[allele_df_key]).
        """
        super().__init__(
            adata=screen,
            repguide_mask=repguide_mask,
            sample_mask_column=sample_mask_column,
            shrink_alpha=shrink_alpha,
            condition_column=condition_column,
            control_condition=control_condition,
            **kwargs,
        )

    def _post_init(
        self,
        allele_df_key: str = None,
        allele_col: str = None,
        alleles_to_select: pd.MultiIndex = None,
        edit_index: pd.DataFrame = None,
        control_guide_tag: str = None,
    ):
        """Called in initialization of child classes to add tiling-screen specific attributes.

        Args
            allele_df_key: A string representing the allele count dataframe key in screen.uns.
            allele_col: A string representing the name of the allele column in the allele count dataframe.
            alleles_to_select: A Pandas multi-index representing the alleles that need to be subsetted.
            edit_index: A Pandas DataFrame that specifies the edit sequences to index.
        """
        if allele_col is None:
            if "allele" in self.screen.uns[allele_df_key].reset_index():
                allele_col = "allele"
            if "aa_allele" in self.screen.uns[allele_df_key].reset_index():
                allele_col = "aa_allele"
        if alleles_to_select is not None:
            allele_counts_selected = (
                self.screen.uns[allele_df_key]
                .set_index(["guide", allele_col])
                .loc[alleles_to_select, :]
                .reset_index()
            )
        else:
            allele_counts_selected = self.screen.uns[allele_df_key]
        if control_guide_tag is not None:

            def _set_uid_to_row(row):
                if control_guide_tag in row.guide:
                    row[allele_col].set_uid(row.guide)

            allele_counts_selected.apply(lambda row: _set_uid_to_row(row), axis=1)
            assert (
                allele_counts_selected[allele_col].map(lambda a: "!" in str(a)).any()
            ), "uid not assinged."
        guide_to_allele, reindexed_df = self.reindex_allele_df(
            allele_counts_selected
        )  # TODO: Fix this for combining different sorting scheme
        self.n_max_alleles = (
            reindexed_df.index.get_level_values("allele_id_for_guide").max() + 1
        )  # include no edit allele

        if edit_index is None:
            self.edit_index = get_edit_to_index_dict(guide_to_allele.aa_allele)
        else:
            self.edit_index = edit_index
        self.n_edits = len(self.edit_index.keys())

        self.allele_to_edit = self.get_allele_to_edit_tensor(
            self.screen_control, self.edit_index, guide_to_allele
        )
        assert self.allele_to_edit.shape == (
            self.n_guides,
            self.n_max_alleles - 1,
            self.n_edits,
        )

        # TODO: fix this for sample-separated data
        self.allele_counts = self.transform_allele(self.screen_selected, reindexed_df)
        assert self.allele_counts.shape == (
            self.n_reps,
            self.n_condits,
            self.n_guides,
            self.n_max_alleles,
        )
        self.allele_counts_control = self.transform_allele_control(
            self.screen_control, reindexed_df
        )
        self.allele_mask = self.get_allele_mask(self.screen_control, guide_to_allele)
        assert self.allele_mask.shape == (self.n_guides, self.n_max_alleles)
        assert self.allele_counts_control.shape == (
            self.n_reps,
            1,
            self.n_guides,
            self.n_max_alleles,
        )

    def get_allele_to_edit_tensor(
        self,
        screen,
        edits_to_index: Dict[str, int],
        guide_allele_id_to_allele_df: pd.DataFrame,
    ) -> torch.Tensor:
        """
        Convert (guide, allele_id_for_guide) -> allele DataFrame into the tensor with shape (n_guides, n_max_alleles_per_guide, n_edits) tensor.

        Args
            edits_to_index: Dictionary from edit (str) to unique index (int)
            guide_allele_id_to_allele_df: pd.DataFrame of (guide(str), allele_id_for_guide(int)) -> CodingNoncodingAllele

        Returns
            allele_edit_assignment: Binary tensor of shape (n_guides, n_max_alleles_per_guide, n_edits. allele_edit_assignment(i, j, k) is 1 if jth allele of ith guide has kth edit.
        """
        guide_allele_id_to_allele_df[
            "edits"
        ] = guide_allele_id_to_allele_df.aa_allele.map(
            lambda a: list(a.aa_allele.edits) + list(a.nt_allele.edits)
        )
        guide_allele_id_to_allele_df = guide_allele_id_to_allele_df.reset_index()
        guide_allele_id_to_allele_df[
            "edit_idx"
        ] = guide_allele_id_to_allele_df.edits.map(
            lambda es: [edits_to_index[e.get_abs_edit()] for e in es]
        )
        guide_allele_id_to_edit_df = guide_allele_id_to_allele_df[
            ["guide", "allele_id_for_guide", "edit_idx"]
        ].set_index(["guide", "allele_id_for_guide"])
        guide_allele_id_to_edit_df = guide_allele_id_to_edit_df.unstack(
            level=1, fill_value=[]
        ).reindex(screen.guides.index, fill_value=[])
        allele_edit_assignment = torch.zeros(
            (len(screen.guides), self.n_max_alleles - 1, len(edits_to_index.keys()))
        )
        for i in range(len(guide_allele_id_to_edit_df)):
            for j in range(len(guide_allele_id_to_edit_df.columns)):
                allele_edit_assignment[i, j, guide_allele_id_to_edit_df.iloc[i, j]] = 1
        return allele_edit_assignment

    def reindex_allele_df(self, alleles_df):
        """
        Input: Dataframe of (guide, allele) -> (per sample count)
        Output:
            - guide_allele_id_to_allele
                DataFrame of (guide, allele_id_for_guide) -> (global_allele_id, aa_allele(str))
            - reindexed_allele_df
                Dataframe of (guide, allele_id_for_guide) -> (per sample count)

            allele_id_for_guide: order of allele within a guide.
            global_allele_id: global unique id for each (guide, allele) pair.
        """
        guide_to_allele = dict(
            list(alleles_df[["guide", "aa_allele"]].groupby("guide").aa_allele)
        )
        dfs = []
        for k, s in guide_to_allele.items():
            df = s.reset_index()
            df = df.rename(columns={"index": "allele_id"})
            df.index = df.index + 1
            df = df.reset_index()
            df = df.rename(columns={"index": "allele_id_for_guide"})
            df["guide"] = k
            dfs.append(df)

        guide_to_allele_tbl = pd.concat(dfs)

        alleles_df = pd.merge(
            alleles_df, guide_to_allele_tbl, on=["aa_allele", "guide"]
        )
        reindexed_df = alleles_df.reset_index().set_index(
            ["guide", "allele_id_for_guide"]
        )
        guide_allele_id_to_allele = reindexed_df[["index", "aa_allele"]]
        reindexed_allele_df = reindexed_df.drop(["aa_allele", "index"], axis=1)
        return (guide_allele_id_to_allele, reindexed_allele_df)

    def transform_allele(self, adata, reindexed_df):
        """
        Transform reindexed allele dataframe reindexed_df of (guide, allele_id_for_guide) -> (per sample count)
        to (n_reps, n_bins, n_guides, n_alleles) tensor.
        """
        allele_tensor = torch.empty(
            (self.n_reps, self.n_condits, self.n_guides, self.n_max_alleles),
        )
        if not self.device is None:
            allele_tensor = allele_tensor.cuda()
        for i in range(self.n_reps):
            for j in range(self.n_condits):
                condit_idx = np.where(
                    (adata.samples.rep_id == i)
                    & (adata.samples[f"{self.condition_column}_id"] == j)
                )[0]
                assert len(condit_idx) == 1, print(i, j, condit_idx)
                condit_idx = condit_idx.item()
                condit_name = adata.samples.index[condit_idx]
                condit_allele_df = (
                    reindexed_df.loc[:, condit_name]
                    .unstack(level=1, fill_value=0)
                    .astype(int)
                )
                condit_allele_df = condit_allele_df.reindex(
                    adata.guides.index, fill_value=0
                )
                condit_bcmatch_counts = adata.layers["X_bcmatch"][:, condit_idx].astype(
                    int
                )
                # if not (condit_bcmatch_counts >= condit_allele_df.sum(axis=1)).all():
                #     print(
                #         f"Allele counts are larger than total bcmatch counts in rep {i}, {j} by {(condit_bcmatch_counts - condit_allele_df.sum(axis=1)).min()}."
                #     )
                condit_allele_df[0] = condit_bcmatch_counts - condit_allele_df.loc[
                    :, condit_allele_df.columns != 0
                ].sum(axis=1)
                condit_allele_df[0].loc[condit_allele_df[0] < 0] = 0
                condit_allele_df = condit_allele_df.sort_values(
                    "allele_id_for_guide", axis=1, ascending=True
                )
                assert all(condit_allele_df.columns == list(range(self.n_max_alleles)))
                assert condit_allele_df.to_numpy().shape == (
                    self.n_guides,
                    self.n_max_alleles,
                )
                allele_tensor[i, j, :, :] = torch.as_tensor(condit_allele_df.to_numpy())

        try:
            assert (allele_tensor >= 0).all(), allele_tensor[allele_tensor < 0]
        except:
            print("Allele tensor doesn't match condit_allele_df")
            return (allele_tensor, reindexed_df)
        return allele_tensor

    def transform_allele_control(self, adata, reindexed_df):
        """
        Transform reindexed allele dataframe reindexed_df of (guide, allele_id_for_guide) -> (per sample count)
        to (n_reps, n_bins, n_guides, n_alleles) tensor.
        """

        allele_tensor = torch.empty(
            (self.n_reps, 1, self.n_guides, self.n_max_alleles),
        )
        if not self.device is None:
            allele_tensor = allele_tensor.cuda()
        for i in range(self.n_reps):
            condit_idx = np.where(adata.samples.rep_id == i)[0]
            assert len(condit_idx) == 1, print(i, self.control_condition, condit_idx)
            condit_idx = condit_idx.item()
            condit_name = adata.samples.index[condit_idx]
            condit_allele_df = (
                reindexed_df.loc[:, condit_name]
                .unstack(level=1, fill_value=0)
                .astype(int)
            )
            condit_allele_df = condit_allele_df.reindex(
                adata.guides.index, fill_value=0
            )
            condit_bcmatch_counts = adata.layers["X_bcmatch"][:, condit_idx].astype(int)
            # if not (condit_bcmatch_counts >= condit_allele_df.sum(axis=1)).all():
            #     print(
            #         "Allele counts are larger than total bcmatch counts in rep {}, .".format(
            #             i,
            #         )
            #     )
            #     print(
            #         f"min value: {(condit_bcmatch_counts - condit_allele_df.sum(axis=1)).min()}"
            #     )
            condit_allele_df[0] = condit_bcmatch_counts - condit_allele_df.loc[
                :, condit_allele_df.columns != 0
            ].sum(axis=1)
            condit_allele_df[0].loc[condit_allele_df[0] < 0] = 0
            condit_allele_df = condit_allele_df.sort_values(
                "allele_id_for_guide", axis=1, ascending=True
            )
            assert all(condit_allele_df.columns == list(range(self.n_max_alleles)))
            assert condit_allele_df.to_numpy().shape == (
                self.n_guides,
                self.n_max_alleles,
            )
            try:
                allele_tensor[i, 0, :, :] = torch.as_tensor(condit_allele_df.to_numpy())
            except:
                print("Allele tensor doesn't match condit_allele_df")
                return (allele_tensor, torch.as_tensor(condit_allele_df.to_numpy()))
        try:
            assert (allele_tensor >= 0).all(), allele_tensor[allele_tensor < 0]
        except:
            print("Negative values in allele_tensor")
            return (allele_tensor, reindexed_df)
        return allele_tensor

    def get_allele_mask(
        self, adata, guide_allele_id_to_allele_df, keep_one_allele: bool = False
    ):
        """
        =====
        Arguments
        keep_one_allele -- If True, if no valid allele exist, still provide mask for one valid allele.
        """
        guide_to_allele_tbl = guide_allele_id_to_allele_df.reset_index()
        n_valid_allele = (
            guide_to_allele_tbl.groupby("guide")["allele_id_for_guide"]
            .max()
            .reindex(adata.guides.index, fill_value=0)
        )
        if keep_one_allele:
            n_valid_allele[n_valid_allele == 0] = 1
        mask = torch.zeros((self.n_guides, self.n_max_alleles))
        for i in range(mask.shape[0]):
            mask[i, 0] = 1
            for j in range(n_valid_allele[i]):
                mask[i, j + 1] = 1
        return mask.bool()

    def get_allele_to_edit_tensor(
        self,
        adata,
        edits_to_index: Dict[str, int],
        guide_allele_id_to_allele_df: pd.DataFrame,
    ):
        """
        Convert (guide, allele_id_for_guide) -> allele DataFrame into the tensor with shape (n_guides, n_max_alleles_per_guide, n_edits) tensor.
        -----
        Arguments
        edits_to_index (dict) -- Dictionary from edit (str) to unique index (int)
        guide_allele_id_to_allele_df (pd.DataFrame) -- pd.DataFrame of (guide(str), allele_id_for_guide(int)) -> CodingNoncodingAllele
        -----
        Returns
        allele_edit_assignment (torch.Tensor) -- Binary tensor of shape (n_guides, n_max_alleles_per_guide, n_edits).
        allele_edit_assignment(i, j, k) is 1 if jth allele of ith guide has kth edit.
        """
        guide_allele_id_to_allele_df[
            "edits"
        ] = guide_allele_id_to_allele_df.aa_allele.map(
            lambda a: list(a.aa_allele.edits) + list(a.nt_allele.edits)
        )
        guide_allele_id_to_allele_df = guide_allele_id_to_allele_df.reset_index()
        guide_allele_id_to_allele_df[
            "edit_idx"
        ] = guide_allele_id_to_allele_df.edits.map(
            lambda es: [edits_to_index[e.get_abs_edit()] for e in es]
        )
        guide_allele_id_to_edit_df = guide_allele_id_to_allele_df[
            ["guide", "allele_id_for_guide", "edit_idx"]
        ].set_index(["guide", "allele_id_for_guide"])
        guide_allele_id_to_edit_df = guide_allele_id_to_edit_df.unstack(
            level=1, fill_value=[]
        ).reindex(adata.guides.index, fill_value=[])
        allele_edit_assignment = torch.zeros(
            (len(adata.guides), self.n_max_alleles - 1, len(edits_to_index.keys()))
        )
        for i in range(len(guide_allele_id_to_edit_df)):
            for j in range(len(guide_allele_id_to_edit_df.columns)):
                allele_edit_assignment[i, j, guide_allele_id_to_edit_df.iloc[i, j]] = 1
        return allele_edit_assignment


@dataclass
class SortingScreenData(ScreenData):
    upper_bounds: torch.Tensor
    lower_bounds: torch.Tensor

    def __init__(
        self,
        screen: be.ReporterScreen,
        repguide_mask: str = None,
        sample_mask_column: str = None,
        shrink_alpha: bool = False,
        condition_column: str = "sort",
        control_condition: str = "bulk",
        lower_quantile_column: str = "lower_quantile",
        upper_quantile_column: str = "upper_quantile",
        replicate_column: str = "rep",
        **kwargs,
    ):
        """
        Args
        ---
        lower_quantile_column: column in screen.samples that indicate lower quantile threshold of sorting bin.
        upper_quantile_column: column in screen.samples that indicate upper quantile threshold of sorting bin.
        """
        self._pre_init(lower_quantile_column, upper_quantile_column, replicate_column)
        self.n_bins = self.n_condits
        super().__init__(
            screen=screen,
            repguide_mask=repguide_mask,
            sample_mask_column=sample_mask_column,
            shrink_alpha=shrink_alpha,
            condition_column=condition_column,
            control_condition=control_condition,
            **kwargs,
        )
        self.condition_column = condition_column

    def _pre_init(
        self,
        lower_quantile_column,
        upper_quantile_column,
    ):
        if (self.screen.samples[lower_quantile_column] < 0.0).any() or (
            self.screen.samples[lower_quantile_column] > 1.0
        ).any():
            raise ValueError(
                f"Invalid quantile value({self.screen.samples[lower_quantile_column]}) in screen.samples[{lower_quantile_column}]: check input."
            )
        if (self.screen.samples[upper_quantile_column] < 0.0).any() or (
            self.screen.samples[upper_quantile_column] > 1.0
        ).any():
            raise ValueError(
                f"Invalid quantile value ({self.screen.samples[upper_quantile_column]}) in screen.samples[{upper_quantile_column}]: check input."
            )
        if (
            self.screen.samples[upper_quantile_column]
            - self.screen.samples[lower_quantile_column]
            < 0
        ).any():
            raise ValueError(
                f"Not all screen.samples[{upper_quantile_column}] larger than screen.samples[{upper_quantile_column}]: check input."
            )

        if not (
            self.screen.samples.groupby(
                [upper_quantile_column, lower_quantile_column]
            ).size()
            == len(self.screen.samples[self.replicate_column].unique())
        ).all():
            raise ValueError(
                "Not all replicate share same quantile bin definition. If you have missing bin data, add the sample and add 'mask' column in 'screen.samples'."
            )
        sorting_bins = self.screen_selected.samples.sort_values(
            [upper_quantile_column, lower_quantile_column]
        )[[upper_quantile_column, lower_quantile_column]].drop_duplicates()
        for j, (idx, row) in enumerate(sorting_bins.iterrows()):
            self.screen_selected.samples.loc[
                (
                    self.screen_selected.samples[upper_quantile_column]
                    == row[upper_quantile_column]
                )
                & (
                    self.screen_selected.samples[lower_quantile_column]
                    == row[lower_quantile_column]
                ),
                f"{self.condition_column}_id",
            ] = j
        self.upper_bounds = torch.as_tensor(sorting_bins[upper_quantile_column].values)
        self.lower_bounds = torch.as_tensor(sorting_bins[lower_quantile_column].values)
        self.screen.samples[f"{self.condition_column}_id"] = -1
        self.screen.samples.loc[
            self.screen_selected.samples.index, f"{self.condition_column}_id"
        ] = self.screen_selected.samples[f"{self.condition_column}_id"]
        self.screen = _assign_rep_ids_and_sort(
            self.screen, self.replicate_column, self.condition_column
        )
        self.screen_selected = _assign_rep_ids_and_sort(
            self.screen_selected, self.replicate_column, self.condition_column
        )
        self.screen_control = _assign_rep_ids_and_sort(
            self.screen_control,
            self.replicate_column,
        )


@dataclass
class SurvivalScreenData(ScreenData):
    def __init__(
        self,
        screen: be.ReporterScreen,
        repguide_mask: str = None,
        sample_mask_column: str = None,
        shrink_alpha: bool = False,
        condition_column: str = "time",
        control_condition: str = "bulk",
        accessibility_col: str = None,
        accessibility_bw_path: str = None,
        **kwargs,
    ):
        super().__init__(
            screen=screen,
            repguide_mask=repguide_mask,
            sample_mask_column=sample_mask_column,
            shrink_alpha=shrink_alpha,
            condition_column=condition_column,
            control_condition=control_condition,
            accessibility_col=accessibility_col,
            accessibility_bw_path=accessibility_bw_path,
            **kwargs,
        )
        self.timepoints = torch.as_tensor(
            self.screen.samples[condition_column].unique()
        )
        self.timepoints = Alias("n_condits")


@dataclass
class VariantReporterScreenData(VariantScreenData, ReporterScreenData):
    def __init__(
        self,
        screen: be.ReporterScreen,
        repguide_mask: str = None,
        sample_mask_column: str = None,
        pi_popt: Tuple[float] = None,
        impute_pi_popt: bool = False,
        shrink_alpha: bool = False,
        condition_column: str = "time",
        control_condition: str = "bulk",
        accessibility_col: str = None,
        accessibility_bw_path: str = None,
        use_const_pi: bool = False,
        pi_prior_count: int = 10,
        **kwargs,
    ):
        VariantScreenData.__init__(
            self,
            screen=screen,
            repguide_mask=repguide_mask,
            sample_mask_column=sample_mask_column,
            condition_column=condition_column,
            control_condition=control_condition,
            accessibility_col=accessibility_col,
            accessibility_bw_path=accessibility_bw_path,
            **kwargs,
        )
        ReporterScreenData._post_init(
            self,
            screen,
            use_const_pi,
            impute_pi_popt,
            pi_prior_count,
            shrink_alpha,
            pi_popt,
        )


@dataclass
class VariantSortingScreenData(VariantScreenData, SortingScreenData):
    def __init__(
        self,
        screen,
        *args,
        lower_quantile_column="lower_quantile",
        upper_quantile_column="upper_quantile",
        replicate_column="rep",
        condition_column="bin",
        **kwargs,
    ):
        SortingScreenData._pre_init(
            self,
            lower_quantile_column,
            upper_quantile_column,
        )
        VariantScreenData.__init__(self, *args, **kwargs)


@dataclass
class VariantSortingReporterScreenData(VariantReporterScreenData, SortingScreenData):
    def __init__(
        self,
        screen,
        *args,
        lower_quantile_column="lower_quantile",
        upper_quantile_column="upper_quantile",
        replicate_column="rep",
        condition_column="bin",
        target_col="target",
        sample_mask_column="mask",
        use_const_pi: bool = False,
        impute_pi_popt: bool = False,
        pi_prior_count: int = 10,
        shrink_alpha: bool = False,
        pi_popt: Tuple[float] = None,
        **kwargs,
    ):
        ScreenData.__init__(
            self,
            screen,
            *args,
            sample_mask_column="mask",
            replicate_column="rep",
            condition_column="bin",
            shrink_alpha=shrink_alpha,
            **kwargs,
        )
        SortingScreenData._pre_init(
            self,
            lower_quantile_column,
            upper_quantile_column,
        )
        ScreenData._post_init(self)
        VariantScreenData._post_init(self, target_col)
        ReporterScreenData._post_init(
            self,
            screen,
            use_const_pi,
            impute_pi_popt,
            pi_prior_count,
            shrink_alpha,
            pi_popt,
        )


@dataclass
class TilingSortingReporterScreenData(TilingReporterScreenData, SortingScreenData):
    def __init__(
        self,
        screen,
        *args,
        lower_quantile_column="lower_quantile",
        upper_quantile_column="upper_quantile",
        replicate_column="rep",
        condition_column="bin",
        sample_mask_column="mask",
        use_const_pi: bool = False,
        impute_pi_popt: bool = False,
        pi_prior_count: int = 10,
        shrink_alpha: bool = False,
        pi_popt: Tuple[float] = None,
        allele_df_key: str = None,
        allele_col: str = None,
        control_guide_tag: str = None,
        **kwargs,
    ):
        ScreenData.__init__(
            self,
            screen,
            *args,
            sample_mask_column="mask",
            replicate_column="rep",
            condition_column="bin",
            shrink_alpha=shrink_alpha,
            **kwargs,
        )
        SortingScreenData._pre_init(
            self,
            lower_quantile_column,
            upper_quantile_column,
        )
        ScreenData._post_init(self)
        TilingReporterScreenData._post_init(
            self,
            allele_df_key=allele_df_key,
            control_guide_tag=control_guide_tag,
        )
        ReporterScreenData._post_init(
            self,
            screen,
            use_const_pi,
            impute_pi_popt,
            pi_prior_count,
            shrink_alpha,
            pi_popt,
        )


@dataclass
class VariantSurvivalReporterScreenData(VariantReporterScreenData, SurvivalScreenData):
    pass



import torch
import pyro
from pyro import poutine
import pyro.distributions as dist
import torch.distributions.constraints as constraints
from .utils import (
    get_alpha,
    get_std_normal_prob,
    scale_pi_by_accessibility,
)
from ..preprocessing.data_class import (
    VariantSortingScreenData,
    VariantSortingReporterScreenData,
    TilingSortingReporterScreenData,
)


def NormalModel(
    data: VariantSortingScreenData, mask_thres: int = 10, use_bcmatch: bool = True
):
    """
    Fit only on guide counts
    Args
    --
    data: input data
    """
    replicate_plate = pyro.plate("rep_plate", data.n_reps, dim=-3)
    replicate_plate2 = pyro.plate("rep_plate2", data.n_reps, dim=-2)
    bin_plate = pyro.plate("bin_plate", data.n_condits, dim=-2)
    guide_plate = pyro.plate("guide_plate", data.n_guides, dim=-1)

    # Set the prior for phenotype means
    with pyro.plate("guide_plate0", 1):
        with pyro.plate("guide_plate1", data.n_targets):
            mu_alleles = pyro.sample("mu_alleles", dist.Laplace(0, 1))
            sd_alleles = pyro.sample(
                "sd_alleles",
                dist.LogNormal(
                    torch.zeros((data.n_targets, 1)), torch.ones(data.n_targets, 1)
                ),
            )
    mu_center = mu_alleles
    mu = torch.repeat_interleave(mu_center, data.target_lengths, dim=0)
    assert mu.shape == (data.n_guides, 1)
    sd = sd_alleles
    sd = torch.repeat_interleave(sd, data.target_lengths, dim=0)
    assert sd.shape == (data.n_guides, 1)

    with replicate_plate:
        with bin_plate as b:
            uq = data.upper_bounds[b]
            lq = data.lower_bounds[b]
            assert uq.shape == lq.shape == (data.n_condits,)
            # with guide_plate, poutine.mask(mask=(data.allele_counts.sum(axis=-1) == 0)):
            with guide_plate:
                alleles_p_bin = get_std_normal_prob(
                    uq.unsqueeze(-1).unsqueeze(-1).expand((-1, data.n_guides, 1)),
                    lq.unsqueeze(-1).unsqueeze(-1).expand((-1, data.n_guides, 1)),
                    mu.unsqueeze(0).expand((data.n_condits, -1, -1)),
                    sd.unsqueeze(0).expand((data.n_condits, -1, -1)),
                )
                assert alleles_p_bin.shape == (data.n_condits, data.n_guides, 1)

            expected_allele_p = alleles_p_bin.unsqueeze(0).expand(
                data.n_reps, -1, -1, -1
            )
            expected_guide_p = expected_allele_p.sum(axis=-1)
            assert expected_guide_p.shape == (
                data.n_reps,
                data.n_condits,
                data.n_guides,
            ), expected_guide_p.shape

    with replicate_plate2:
        with pyro.plate("guide_plate3", data.n_guides, dim=-1):
            a = get_alpha(expected_guide_p, data.size_factor, data.sample_mask, data.a0)
            a_bcmatch = get_alpha(
                expected_guide_p,
                data.size_factor_bcmatch,
                data.sample_mask,
                data.a0_bcmatch,
            )
            assert (
                data.X.shape
                == data.X_bcmatch.shape
                == (
                    data.n_reps,
                    data.n_condits,
                    data.n_guides,
                )
            )
            with poutine.mask(
                mask=torch.logical_and(
                    data.X_masked.permute(0, 2, 1).sum(axis=-1) > mask_thres,
                    data.repguide_mask,
                )
            ):
                pyro.sample(
                    "guide_counts",
                    dist.DirichletMultinomial(a, validate_args=False),
                    obs=data.X_masked.permute(0, 2, 1),
                )
            if use_bcmatch:
                with poutine.mask(
                    mask=torch.logical_and(
                        data.X_bcmatch_masked.permute(0, 2, 1).sum(axis=-1)
                        > mask_thres,
                        data.repguide_mask,
                    )
                ):
                    assert (
                        a_bcmatch > 0
                    ).all(), f"{torch.where(a_bcmatch < 1e-6)}\n{torch.where(torch.isnan(a_bcmatch))}"
                    pyro.sample(
                        "guide_bcmatch_counts",
                        dist.DirichletMultinomial(a_bcmatch, validate_args=False),
                        obs=data.X_bcmatch_masked.permute(0, 2, 1),
                    )

    return alleles_p_bin


def ControlNormalModel(data, mask_thres=10, use_bcmatch=True):
    """
    Fit shared control distribution
    """
    replicate_plate = pyro.plate("rep_plate", data.n_reps, dim=-3)
    replicate_plate2 = pyro.plate("rep_plate2", data.n_reps, dim=-2)
    bin_plate = pyro.plate("bin_plate", data.n_condits, dim=-2)
    guide_plate = pyro.plate("guide_plate", data.n_guides, dim=-1)

    # Set the prior for phenotype means
    mu_alleles = pyro.sample("mu_alleles", dist.Laplace(0, 1))
    sd_alleles = pyro.sample("sd_alleles", dist.LogNormal(0, 1))
    mu = mu_alleles.repeat(data.n_guides).unsqueeze(-1)
    sd = sd_alleles.repeat(data.n_guides).unsqueeze(-1)

    with replicate_plate:
        with bin_plate as b:
            uq = data.upper_bounds[b]
            lq = data.lower_bounds[b]
            assert uq.shape == lq.shape == (data.n_condits,)
            with guide_plate:
                alleles_p_bin = get_std_normal_prob(
                    uq.unsqueeze(-1).unsqueeze(-1).expand((-1, data.n_guides, 1)),
                    lq.unsqueeze(-1).unsqueeze(-1).expand((-1, data.n_guides, 1)),
                    mu.unsqueeze(0).expand((data.n_condits, -1, -1)),
                    sd.unsqueeze(0).expand((data.n_condits, -1, -1)),
                )
                assert alleles_p_bin.shape == (data.n_condits, data.n_guides, 1)

            expected_allele_p = alleles_p_bin.unsqueeze(0).expand(
                data.n_reps, -1, -1, -1
            )
            expected_guide_p = expected_allele_p.sum(axis=-1)
            assert expected_guide_p.shape == (
                data.n_reps,
                data.n_condits,
                data.n_guides,
            ), expected_guide_p.shape

    with replicate_plate2:
        with pyro.plate("guide_plate3", data.n_guides, dim=-1):
            a = get_alpha(expected_guide_p, data.size_factor, data.sample_mask, data.a0)
            a_bcmatch = get_alpha(
                expected_guide_p,
                data.size_factor_bcmatch,
                data.sample_mask,
                data.a0_bcmatch,
            )
            assert (
                data.X.shape
                == data.X_bcmatch.shape
                == (
                    data.n_reps,
                    data.n_condits,
                    data.n_guides,
                )
            )
            with poutine.mask(
                mask=torch.logical_and(
                    data.X_masked.permute(0, 2, 1).sum(axis=-1) > mask_thres,
                    data.repguide_mask,
                )
            ):
                pyro.sample(
                    "guide_counts",
                    dist.DirichletMultinomial(a, validate_args=False),
                    obs=data.X_masked.permute(0, 2, 1),
                )
            if use_bcmatch:
                with poutine.mask(
                    mask=torch.logical_and(
                        data.X_bcmatch_masked.permute(0, 2, 1).sum(axis=-1)
                        > mask_thres,
                        data.repguide_mask,
                    )
                ):
                    pyro.sample(
                        "guide_bcmatch_counts",
                        dist.DirichletMultinomial(a_bcmatch, validate_args=False),
                        obs=data.X_bcmatch_masked.permute(0, 2, 1),
                    )

    return alleles_p_bin


def MixtureNormalConstPiModel(
    data: VariantSortingScreenData,
    alpha_prior: float = 1,
    use_bcmatch: bool = True,
    sd_scale: float = 0.01,
):
    """
    model B0 + proper pi + Multinomial counts
    Args:
        scale_by_accessibility: If True, pi fitted from reporter data is scaled by accessibility*0.128
    """
    replicate_plate = pyro.plate("rep_plate", data.n_reps, dim=-3)
    replicate_plate2 = pyro.plate("rep_plate2", data.n_reps, dim=-2)
    bin_plate = pyro.plate("bin_plate", data.n_condits, dim=-2)
    guide_plate = pyro.plate("guide_plate", data.n_guides, dim=-1)

    # Set the prior for phenotype means
    with pyro.plate("guide_plate0", 1):
        with pyro.plate("guide_plate1", data.n_targets):
            mu_alleles = pyro.sample("mu_alleles", dist.Laplace(0, 1))
            sd_alleles = pyro.sample(
                "sd_alleles",
                dist.LogNormal(
                    torch.zeros((data.n_targets, 1)),
                    torch.ones(data.n_targets, 1) * sd_scale,
                ),
            )
    mu_center = torch.cat([torch.zeros((data.n_targets, 1)), mu_alleles], axis=-1)
    mu = torch.repeat_interleave(mu_center, data.target_lengths, dim=0)
    assert mu.shape == (data.n_guides, 2)

    sd = torch.cat([torch.ones((data.n_targets, 1)), sd_alleles], axis=-1)
    sd = torch.repeat_interleave(sd, data.target_lengths, dim=0)
    assert sd.shape == (data.n_guides, 2)
    # The pi should be Dirichlet distributed instead of independent betas

    pi_a_scaled = data.pi / data.pi.sum(axis=-1)[:, None] * data.pi_a0[:, None]
    with replicate_plate:
        with guide_plate:
            # Accounting for sample specific overall edit rate across all guides.
            # P(allele | guide, bin=bulk)
            pi = pyro.sample(
                "pi",
                dist.Dirichlet(
                    pi_a_scaled.unsqueeze(0).unsqueeze(0).expand(data.n_reps, 1, -1, -1)
                ),
            )
            assert pi.shape == (
                data.n_reps,
                1,
                data.n_guides,
                2,
            ), pi.shape
    with replicate_plate:
        with bin_plate as b:
            uq = data.upper_bounds[b]
            lq = data.lower_bounds[b]
            assert uq.shape == lq.shape == (data.n_condits,)
            # with guide_plate, poutine.mask(mask=(data.allele_counts.sum(axis=-1) == 0)):
            with guide_plate, poutine.mask(mask=data.repguide_mask.unsqueeze(1)):
                alleles_p_bin = get_std_normal_prob(
                    uq.unsqueeze(-1).unsqueeze(-1).expand((-1, data.n_guides, 2)),
                    lq.unsqueeze(-1).unsqueeze(-1).expand((-1, data.n_guides, 2)),
                    mu.unsqueeze(0).expand((data.n_condits, -1, -1)),
                    sd.unsqueeze(0).expand((data.n_condits, -1, -1)),
                )
                assert alleles_p_bin.shape == (data.n_condits, data.n_guides, 2)

            expected_allele_p = (
                pi.expand(data.n_reps, data.n_condits, -1, -1)
                * alleles_p_bin[None, :, :, :]
            )
            expected_guide_p = expected_allele_p.sum(axis=-1)
            assert expected_guide_p.shape == (
                data.n_reps,
                data.n_condits,
                data.n_guides,
            ), expected_guide_p.shape

    with replicate_plate2:
        with pyro.plate("guide_plate3", data.n_guides, dim=-1):
            a = get_alpha(expected_guide_p, data.size_factor, data.sample_mask, data.a0)
            a_bcmatch = get_alpha(
                expected_guide_p,
                data.size_factor_bcmatch,
                data.sample_mask,
                data.a0_bcmatch,
            )
            # a_bcmatch = get_alpha(expected_guide_p, data.size_factor_bcmatch, data.sample_mask, data.a0_bcmatch)
            # assert a.shape == a_bcmatch.shape == (data.n_reps, data.n_guides, data.n_condits)
            assert (
                data.X.shape
                == data.X_bcmatch.shape
                == (
                    data.n_reps,
                    data.n_condits,
                    data.n_guides,
                )
            )
            with poutine.mask(
                mask=torch.logical_and(
                    data.X_masked.permute(0, 2, 1).sum(axis=-1) > 10, data.repguide_mask
                )
            ):
                pyro.sample(
                    "guide_counts",
                    dist.Multinomial(probs=a, validate_args=False),
                    obs=data.X_masked.permute(0, 2, 1),
                )
            if use_bcmatch:
                with poutine.mask(
                    mask=torch.logical_and(
                        data.X_bcmatch_masked.permute(0, 2, 1).sum(axis=-1) > 10,
                        data.repguide_mask,
                    )
                ):
                    pyro.sample(
                        "guide_bcmatch_counts",
                        dist.Multinomial(probs=a_bcmatch, validate_args=False),
                        obs=data.X_bcmatch_masked.permute(0, 2, 1),
                    )


def MixtureNormalModel(
    data: VariantSortingReporterScreenData,
    alpha_prior: float = 1,
    use_bcmatch: bool = True,
    sd_scale: float = 0.01,
    scale_by_accessibility: bool = False,
    fit_noise: bool = False,
):
    """
    Args:
        scale_by_accessibility: If True, pi fitted from reporter data is scaled by accessibility
    """
    torch.autograd.set_detect_anomaly(True)
    replicate_plate = pyro.plate("rep_plate", data.n_reps, dim=-3)
    replicate_plate2 = pyro.plate("rep_plate2", data.n_reps, dim=-2)
    bin_plate = pyro.plate("bin_plate", data.n_condits, dim=-2)
    guide_plate = pyro.plate("guide_plate", data.n_guides, dim=-1)

    # Set the prior for phenotype means
    with pyro.plate("guide_plate0", 1):
        with pyro.plate("guide_plate1", data.n_targets):
            mu_alleles = pyro.sample("mu_alleles", dist.Laplace(0, 1))
            sd_alleles = pyro.sample(
                "sd_alleles",
                dist.LogNormal(
                    torch.zeros((data.n_targets, 1)),
                    torch.ones(data.n_targets, 1) * sd_scale,
                ),
            )
    mu_center = torch.cat([torch.zeros((data.n_targets, 1)), mu_alleles], axis=-1)
    mu = torch.repeat_interleave(mu_center, data.target_lengths, dim=0)
    assert mu.shape == (data.n_guides, 2)

    sd = torch.cat([torch.ones((data.n_targets, 1)), sd_alleles], axis=-1)
    sd = torch.repeat_interleave(sd, data.target_lengths, dim=0)
    assert sd.shape == (data.n_guides, 2)
    # The pi should be Dirichlet distributed instead of independent betas
    alpha_pi = pyro.param(
        "alpha_pi",
        torch.ones(
            (
                data.n_guides,
                2,
            )
        )
        * alpha_prior,
        constraint=constraints.positive,
    )
    pi_a_scaled = alpha_pi / alpha_pi.sum(axis=-1)[:, None] * data.pi_a0[:, None]
    assert alpha_pi.shape == (
        data.n_guides,
        2,
    ), alpha_pi.shape
    with replicate_plate:
        with guide_plate, poutine.mask(mask=data.repguide_mask.unsqueeze(1)):
            # Accounting for sample specific overall edit rate across all guides.
            # P(allele | guide, bin=bulk)
            pi = pyro.sample(
                "pi",
                dist.Dirichlet(
                    pi_a_scaled.unsqueeze(0).unsqueeze(0).expand(data.n_reps, 1, -1, -1)
                ),
            )
            assert pi.shape == (
                data.n_reps,
                1,
                data.n_guides,
                2,
            ), pi.shape
            pyro.sample(
                "bulk_allele_count",
                dist.Multinomial(probs=pi, validate_args=False),
                obs=data.allele_counts_control,
            )
    if scale_by_accessibility:
        # Endogenous target site editing rate may be different
        pi = scale_pi_by_accessibility(
            pi, data.guide_accessibility, fit_noise=fit_noise
        )
    with replicate_plate:
        with bin_plate as b:
            uq = data.upper_bounds[b]
            lq = data.lower_bounds[b]
            assert uq.shape == lq.shape == (data.n_condits,)
            # with guide_plate, poutine.mask(mask=(data.allele_counts.sum(axis=-1) == 0)):
            with guide_plate, poutine.mask(mask=data.repguide_mask.unsqueeze(1)):
                alleles_p_bin = get_std_normal_prob(
                    uq.unsqueeze(-1).unsqueeze(-1).expand((-1, data.n_guides, 2)),
                    lq.unsqueeze(-1).unsqueeze(-1).expand((-1, data.n_guides, 2)),
                    mu.unsqueeze(0).expand((data.n_condits, -1, -1)),
                    sd.unsqueeze(0).expand((data.n_condits, -1, -1)),
                )
                assert alleles_p_bin.shape == (data.n_condits, data.n_guides, 2)

            expected_allele_p = (
                pi.expand(data.n_reps, data.n_condits, -1, -1)
                * alleles_p_bin[None, :, :, :]
            )
            expected_guide_p = expected_allele_p.sum(axis=-1)
            assert expected_guide_p.shape == (
                data.n_reps,
                data.n_condits,
                data.n_guides,
            ), expected_guide_p.shape

    with replicate_plate2:
        with pyro.plate("guide_plate3", data.n_guides, dim=-1):
            a = get_alpha(expected_guide_p, data.size_factor, data.sample_mask, data.a0)
            a_bcmatch = get_alpha(
                expected_guide_p,
                data.size_factor_bcmatch,
                data.sample_mask,
                data.a0_bcmatch,
            )
            # a_bcmatch = get_alpha(expected_guide_p, data.size_factor_bcmatch, data.sample_mask, data.a0_bcmatch)
            # assert a.shape == a_bcmatch.shape == (data.n_reps, data.n_guides, data.n_condits)
            assert (
                data.X.shape
                == data.X_bcmatch.shape
                == (
                    data.n_reps,
                    data.n_condits,
                    data.n_guides,
                )
            )
            with poutine.mask(
                mask=torch.logical_and(
                    data.X_masked.permute(0, 2, 1).sum(axis=-1) > 10, data.repguide_mask
                )
            ):
                pyro.sample(
                    "guide_counts",
                    dist.DirichletMultinomial(a, validate_args=False),
                    obs=data.X_masked.permute(0, 2, 1),
                )
            if use_bcmatch:
                with poutine.mask(
                    mask=torch.logical_and(
                        data.X_bcmatch_masked.permute(0, 2, 1).sum(axis=-1) > 10,
                        data.repguide_mask,
                    )
                ):
                    pyro.sample(
                        "guide_bcmatch_counts",
                        dist.DirichletMultinomial(a_bcmatch, validate_args=False),
                        obs=data.X_bcmatch_masked.permute(0, 2, 1),
                    )


def NormalGuide(data):
    with pyro.plate("guide_plate0", 1):
        with pyro.plate("guide_plate1", data.n_targets):
            mu_loc = pyro.param("mu_loc", torch.zeros((data.n_targets, 1)))
            mu_scale = pyro.param(
                "mu_scale",
                torch.ones((data.n_targets, 1)),
                constraint=constraints.positive,
            )
            pyro.sample("mu_alleles", dist.Normal(mu_loc, mu_scale))
            sd_loc = pyro.param("sd_loc", torch.zeros((data.n_targets, 1)))
            sd_scale = pyro.param(
                "sd_scale",
                torch.ones((data.n_targets, 1)),
                constraint=constraints.positive,
            )
            pyro.sample("sd_alleles", dist.LogNormal(sd_loc, sd_scale))


def MixtureNormalGuide(
    data,
    alpha_prior: float = 1,
    use_bcmatch: bool = True,
    scale_by_accessibility: bool = False,
    fit_noise: bool = False,
):
    """
    Guide for MixtureNormal model
    """

    replicate_plate = pyro.plate("rep_plate", data.n_reps, dim=-3)
    guide_plate = pyro.plate("guide_plate", data.n_guides, dim=-1)

    # Set the prior for phenotype means
    mu_loc = pyro.param("mu_loc", torch.zeros((data.n_targets, 1)))
    mu_scale = pyro.param(
        "mu_scale", torch.ones((data.n_targets, 1)), constraint=constraints.positive
    )
    sd_loc = pyro.param("sd_loc", torch.zeros((data.n_targets, 1)))
    sd_scale = pyro.param(
        "sd_scale", torch.ones((data.n_targets, 1)), constraint=constraints.positive
    )
    with pyro.plate("guide_plate0", 1):
        with pyro.plate("guide_plate1", data.n_targets):
            mu_alleles = pyro.sample("mu_alleles", dist.Normal(mu_loc, mu_scale))
            sd_alleles = pyro.sample("sd_alleles", dist.LogNormal(sd_loc, sd_scale))
    mu_center = torch.cat([torch.zeros((data.n_targets, 1)), mu_alleles], axis=-1)
    mu = torch.repeat_interleave(mu_center, data.target_lengths, dim=0)
    assert mu.shape == (data.n_guides, 2)

    sd = torch.cat([torch.ones((data.n_targets, 1)), sd_alleles], axis=-1)
    sd = torch.repeat_interleave(sd, data.target_lengths, dim=0)
    assert sd.shape == (data.n_guides, 2)
    # The pi should be Dirichlet distributed instead of independent betas
    alpha_pi = pyro.param(
        "alpha_pi",
        torch.ones(
            (
                data.n_guides,
                2,
            )
        )
        * alpha_prior,
        constraint=constraints.positive,
    )
    assert alpha_pi.shape == (
        data.n_guides,
        2,
    ), alpha_pi.shape
    pi_a_scaled = alpha_pi / alpha_pi.sum(axis=-1)[:, None] * data.pi_a0[:, None]

    with replicate_plate:
        with guide_plate:
            pi = pyro.sample(
                "pi",
                dist.Dirichlet(
                    pi_a_scaled.unsqueeze(0)
                    .unsqueeze(0)
                    .expand(data.n_reps, 1, -1, -1)
                    .clamp(1e-5)
                ),
            )
            assert pi.shape == (
                data.n_reps,
                1,
                data.n_guides,
                2,
            ), pi.shape
    if scale_by_accessibility:
        # Endogenous target site editing rate may be different
        pi = scale_pi_by_accessibility(
            pi, data.guide_accessibility, fit_noise=fit_noise
        )


def ControlNormalGuide(data, mask_thres=10, use_bcmatch=True):
    """
    Fit shared mean
    """
    # Set the prior for phenotype means
    mu_loc = pyro.param("mu_loc", torch.tensor(0.0))
    mu_scale = pyro.param(
        "mu_scale", torch.tensor(1.0), constraint=constraints.positive
    )
    sd_loc = pyro.param("sd_loc", torch.tensor(0.0))
    sd_scale = pyro.param(
        "sd_scale", torch.tensor(1.0), constraint=constraints.positive
    )
    pyro.sample("mu_alleles", dist.Normal(mu_loc, mu_scale))
    pyro.sample("sd_alleles", dist.LogNormal(sd_loc, sd_scale))


def MultiMixtureNormalModel(
    data: TilingSortingReporterScreenData,
    alpha_prior=1,
    use_bcmatch=True,
    sd_scale=0.01,
    norm_pi=False,
    scale_by_accessibility=False,
    epsilon=1e-5,
    fit_noise: bool = False,
):
    """Tiling version of MixtureNormalModel"""

    replicate_plate = pyro.plate("rep_plate", data.n_reps, dim=-3)
    replicate_plate2 = pyro.plate("rep_plate2", data.n_reps, dim=-2)
    bin_plate = pyro.plate("bin_plate", data.n_condits, dim=-2)
    guide_plate = pyro.plate("guide_plate", data.n_guides, dim=-1)

    # Set the prior for phenotype means
    with pyro.plate("guide_plate1", data.n_edits):
        mu_edits = pyro.sample("mu_alleles", dist.Laplace(0, 1))
        sd_edits = pyro.sample(
            "sd_alleles",
            dist.LogNormal(
                torch.zeros((data.n_edits,)),
                torch.ones(
                    data.n_edits,
                )
                * sd_scale,
            ),
        )
    assert mu_edits.shape == sd_edits.shape == (data.n_edits,)
    assert data.allele_to_edit.shape == (
        data.n_guides,
        data.n_max_alleles - 1,
        data.n_edits,
    )
    mu_alleles = torch.matmul(data.allele_to_edit, mu_edits)
    assert mu_alleles.shape == (data.n_guides, data.n_max_alleles - 1)
    sd_alleles = torch.linalg.norm(
        data.allele_to_edit * sd_edits[None, None, :], dim=-1
    )  # Frobenius 2-norm

    mu = torch.cat([torch.zeros((data.n_guides, 1)), mu_alleles], axis=-1)
    sd = torch.cat([torch.ones((data.n_guides, 1)), sd_alleles], axis=-1)
    assert mu.shape == sd.shape == (data.n_guides, data.n_max_alleles), (
        mu.shape,
        sd.shape,
    )

    # The pi should be Dirichlet distributed instead of independent betas
    alpha_pi0 = (
        torch.ones(
            (
                data.n_guides,
                data.n_max_alleles,
            )
        )
        * alpha_prior
    )
    # Effectively remove alphas for non-existing alleles
    assert data.allele_mask.shape == (data.n_guides, data.n_max_alleles)
    alpha_pi0[~data.allele_mask] = epsilon
    alpha_pi = pyro.param("alpha_pi", alpha_pi0, constraint=constraints.positive)
    alpha_pi[~data.allele_mask] = epsilon
    pi_a_scaled = (
        (alpha_pi + epsilon / alpha_pi.shape[-1])
        / (alpha_pi.sum(axis=-1)[:, None] + epsilon)
        * data.pi_a0[:, None]
    )
    pi_a_scaled[pi_a_scaled < epsilon] = epsilon
    if torch.isnan(pi_a_scaled).any():
        print(torch.where(alpha_pi.isnan()))
        print(torch.where(alpha_pi < 0))
        exit(1)
    if (pi_a_scaled <= 0).any():
        print(torch.where(alpha_pi < 0))
    with replicate_plate:
        with guide_plate, poutine.mask(mask=data.repguide_mask.unsqueeze(1)):
            pi = pyro.sample(
                "pi",
                dist.Dirichlet(
                    pi_a_scaled.unsqueeze(0).unsqueeze(0).expand(data.n_reps, 1, -1, -1)
                ),
            )
            pyro.sample(
                "bulk_allele_count",
                dist.Multinomial(probs=pi, validate_args=False),
                obs=data.allele_counts_control,
            )
    if scale_by_accessibility:
        # Endogenous target site editing rate may be different
        pi = scale_pi_by_accessibility(
            pi, data.guide_accessibility, fit_noise=fit_noise
        )

    with replicate_plate:
        with bin_plate as b:
            uq = data.upper_bounds[b]
            lq = data.lower_bounds[b]
            assert uq.shape == lq.shape == (data.n_condits,)
            with guide_plate, poutine.mask(mask=data.repguide_mask.unsqueeze(1)):
                alleles_p_bin = get_std_normal_prob(
                    uq.unsqueeze(-1)
                    .unsqueeze(-1)
                    .expand((-1, data.n_guides, data.n_max_alleles)),
                    lq.unsqueeze(-1)
                    .unsqueeze(-1)
                    .expand((-1, data.n_guides, data.n_max_alleles)),
                    mu.unsqueeze(0).expand((data.n_condits, -1, -1)),
                    sd.unsqueeze(0).expand((data.n_condits, -1, -1)),
                    mask=data.allele_mask.unsqueeze(0).expand((data.n_condits, -1, -1)),
                )
                assert alleles_p_bin.shape == (
                    data.n_condits,
                    data.n_guides,
                    data.n_max_alleles,
                )
            expected_allele_p = (
                pi.expand(data.n_reps, data.n_condits, -1, -1)
                * alleles_p_bin[None, :, :, :]
            )
            expected_guide_p = expected_allele_p.sum(axis=-1)
            assert expected_guide_p.shape == (
                data.n_reps,
                data.n_condits,
                data.n_guides,
            ), expected_guide_p.shape

    with replicate_plate2:
        with pyro.plate("guide_plate3", data.n_guides, dim=-1):
            a = get_alpha(expected_guide_p, data.size_factor, data.sample_mask, data.a0)
            a_bcmatch = get_alpha(
                expected_guide_p,
                data.size_factor_bcmatch,
                data.sample_mask,
                data.a0_bcmatch,
            )
            # a_bcmatch = get_alpha(expected_guide_p, data.size_factor_bcmatch, data.sample_mask, data.a0_bcmatch)
            # assert a.shape == a_bcmatch.shape == (data.n_reps, data.n_guides, data.n_condits)
            assert (
                data.X.shape
                == data.X_bcmatch.shape
                == (
                    data.n_reps,
                    data.n_condits,
                    data.n_guides,
                )
            )
            with poutine.mask(
                mask=torch.logical_and(
                    data.X_masked.permute(0, 2, 1).sum(axis=-1) > 10, data.repguide_mask
                )
            ):
                pyro.sample(
                    "guide_counts",
                    dist.DirichletMultinomial(a, validate_args=False),
                    obs=data.X_masked.permute(0, 2, 1),
                )
            if use_bcmatch:
                with poutine.mask(
                    mask=torch.logical_and(
                        data.X_bcmatch_masked.permute(0, 2, 1).sum(axis=-1) > 10,
                        data.repguide_mask,
                    )
                ):
                    pyro.sample(
                        "guide_bcmatch_counts",
                        dist.DirichletMultinomial(a_bcmatch, validate_args=False),
                        obs=data.X_bcmatch_masked.permute(0, 2, 1),
                    )


def MultiMixtureNormalGuide(
    data,
    alpha_prior=1,
    use_bcmatch=True,
    epsilon=1e-5,
    scale_by_accessibility: bool = False,
    fit_noise: bool = False,
):
    """
    Guide for model C14
    """
    replicate_plate = pyro.plate("rep_plate", data.n_reps, dim=-3)
    guide_plate = pyro.plate("guide_plate", data.n_guides, dim=-1)

    # Set the prior for phenotype means
    mu_loc = pyro.param("mu_loc", torch.zeros((data.n_edits,)))
    mu_scale = pyro.param(
        "mu_scale", torch.ones((data.n_edits,)), constraint=constraints.positive
    )
    sd_loc = pyro.param("sd_loc", torch.zeros((data.n_edits,)))
    sd_scale = pyro.param(
        "sd_scale", torch.ones((data.n_edits,)), constraint=constraints.positive
    )
    with pyro.plate("guide_plate1", data.n_edits):
        mu_edits = pyro.sample("mu_alleles", dist.Normal(mu_loc, mu_scale))
        sd_edits = pyro.sample(
            "sd_alleles",
            dist.LogNormal(sd_loc, sd_scale),
        )
    mu_alleles = torch.matmul(data.allele_to_edit, mu_edits)
    assert mu_alleles.shape == (data.n_guides, data.n_max_alleles - 1), (
        mu_alleles.shape,
        data.n_max_alleles,
        data.n_edits,
    )
    sd_alleles = torch.linalg.norm(
        data.allele_to_edit * sd_edits[None, None, :], dim=-1
    )  # Frobenius 2-norm

    mu = torch.cat([torch.zeros((data.n_guides, 1)), mu_alleles], axis=-1)
    sd = torch.cat([torch.ones((data.n_guides, 1)), sd_alleles], axis=-1)
    assert mu.shape == sd.shape == (data.n_guides, data.n_max_alleles), (
        mu.shape,
        sd.shape,
    )
    # The pi should be Dirichlet distributed instead of independent betas
    alpha_pi0 = (
        torch.ones(
            (
                data.n_guides,
                data.n_max_alleles,
            )
        )
        * alpha_prior
    )
    # Effectively remove alphas for non-existing alleles
    assert data.allele_mask.shape == (data.n_guides, data.n_max_alleles)
    alpha_pi0[~data.allele_mask] = epsilon
    alpha_pi = pyro.param("alpha_pi", alpha_pi0, constraint=constraints.positive)
    alpha_pi[~data.allele_mask] = epsilon
    pi_a_scaled = alpha_pi / alpha_pi.sum(axis=-1)[:, None] * data.pi_a0[:, None]

    with replicate_plate:
        with guide_plate, poutine.mask(mask=data.repguide_mask.unsqueeze(1)):
            pi = pyro.sample(
                "pi",
                dist.Dirichlet(
                    pi_a_scaled.unsqueeze(0)
                    .unsqueeze(0)
                    .expand(data.n_reps, 1, -1, -1)
                    # .clamp(1e-5)
                ),
            )

            assert pi.shape == (
                data.n_reps,
                1,
                data.n_guides,
                data.n_max_alleles,
            ), pi.shape
    if scale_by_accessibility:
        # Endogenous target site editing rate may be different
        pi = scale_pi_by_accessibility(
            pi, data.guide_accessibility, fit_noise=fit_noise
        )



from __future__ import annotations
from typing import Iterable
from enum import IntEnum
import warnings
import numpy as np
import re
from ..framework.Edit import Allele, Edit
from ..utils.arithmetric import jaccard

AA_SET = {
    "A",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "I",
    "K",
    "L",
    "M",
    "N",
    "P",
    "Q",
    "R",
    "S",
    "T",
    "V",
    "W",
    "Y",
    "*",
    "/",
}


class MutationType(IntEnum):
    NO_CHANGE = -1
    SYNONYMOUS = 0
    MISSENSE = 1
    NONSENSE = 2


class AminoAcidEdit(Edit):
    def __init__(self, pos: int, ref: str, alt: str):
        self.pos = int(pos)
        assert ref in AA_SET, "Invalid ref aa: {}".format(ref)
        assert alt in AA_SET, "Invalid alt aa: {}".format(alt)
        self.ref = ref
        self.alt = alt

    @classmethod
    def from_str(cls, edit_str):
        pos, aa_change = edit_str.split(":")
        ref, alt = aa_change.split(">")
        return cls(pos, ref, alt)

    @classmethod
    def match_str(cls, edit_str):
        if isinstance(edit_str, AminoAcidEdit):
            return True
        if edit_str == "":
            return True
        pattern = r"-?\d+:[+-]:[A-Z*-]>[A-Z*-]"
        return re.fullmatch(pattern, edit_str)

    def get_abs_edit(self):
        return "A" + self.__repr__()

    def __repr__(self):
        return "{}:{}>{}".format(int(self.pos), self.ref, self.alt)

    def __hash__(self):
        return hash(self.__repr__())

    def _severity(self):
        if self.ref == self.alt:
            return MutationType.SYNONYMOUS
        if self.alt == "*":
            return MutationType.NONSENSE
        if self.alt in AA_SET:
            return MutationType.MISSENSE
        else:
            raise ValueError("Alt base invalid:{}".format(self.alt))

    def __eq__(self, other):
        if self.pos == other.pos and self.ref == other.ref and self.alt == other.alt:
            return True
        return False

    def __lt__(self, other):  # Implemented for pandas compatibility
        return self.pos < other.pos

    def __gt__(self, other):  # Implemented for pandas compatibility
        return self.pos > other.pos


class AminoAcidAllele(Allele):
    def __init__(self, edits: Iterable[AminoAcidEdit] = None):
        if edits is None:
            self.edits = set()
        else:
            self.edits = set(edits)

    @classmethod
    def match_str(cls, allele_str):
        if isinstance(allele_str, AminoAcidAllele):
            return True
        # try:
        return all(map(AminoAcidEdit.match_str, allele_str.split(",")))

    @classmethod
    def from_str(cls, allele_str):  # pos:strand:start>end
        if type(allele_str) is AminoAcidAllele:
            return allele_str
        edits = set()
        try:
            for edit_str in allele_str.split(","):
                edit = AminoAcidEdit.from_str(edit_str)
                edits.add(edit)
        except ValueError:
            if allele_str.strip() == "":
                return cls(None)
        return cls(edits)

    def get_most_severe(self):
        if len(self.edits) == 0:
            return MutationType.NO_CHANGE
        severities = []
        for edit in self.edits:
            severities.append(edit._severity())
        return max(severities)

    def get_most_severe_edit(self):
        if len(self.edits) == 0:
            return None
        severities = []
        edits = []
        for edit in self.edits:
            severities.append(edit._severity())
            edits.append(edit)
        return edits[np.argmax(severities)]


class CodingNoncodingAllele(Allele):
    def __init__(
        self,
        aa_edits: Iterable[AminoAcidEdit] = None,
        base_edits: Iterable[Edit] = None,
        unique_identifier=None,
    ):
        self.aa_allele = AminoAcidAllele(aa_edits)
        self.nt_allele = Allele(base_edits)
        self.uid = unique_identifier
        if not self.uid is None:
            self.set_uid(self.uid)

    @classmethod
    def from_str(cls, allele_str):
        if isinstance(allele_str, cls):
            return allele_str
        try:
            aa_allele_str, nt_allele_str = allele_str.split("|")
            aa_allele = AminoAcidAllele.from_str(aa_allele_str)
            nt_allele = Allele.from_str(nt_allele_str)
        except ValueError as e:
            print(e)
            if allele_str.strip() == "":
                return cls(None)
            else:
                print(allele_str)
                exit(1)
        uid = None
        if (
            len(nt_allele) > 0
            and all([not e.uid is None for e in nt_allele.edits])
            and len(np.unique([e.uid for e in nt_allele.edits if not e.uid is None]))
        ):
            uid = next(iter(nt_allele.edits)).uid
        return cls(aa_allele.edits, nt_allele.edits, uid)

    @classmethod
    def match_str(cls, allele_str):
        if isinstance(allele_str, CodingNoncodingAllele):
            return True
        try:
            aa_allele, nt_allele = allele_str.split("|")
            aa_match = AminoAcidAllele.match_str(aa_allele)
            nt_match = Allele.match_str(nt_allele)
            return aa_match and nt_match
        except:
            return False

    def get_most_severe(self):
        aa_sev = self.aa_allele.get_most_severe()
        if len(self.nt_allele.edits) > 0:
            return max(aa_sev, 0.1)
        return aa_sev

    def get_most_severe_edit(self):
        aa_sev = self.aa_allele.get_most_severe()
        sev_aa = self.aa_allele.get_most_severe_edit()
        if len(self.nt_allele.edits) > 0 and aa_sev > 0.1:
            return next(iter(self.nt_allele.edits))
        return sev_aa

    def set_uid(self, uid):
        self.uid = uid
        self.nt_allele.edits = set([e.set_uid(uid) for e in self.nt_allele.edits])
        # self.aa_allele.edits = set([e.set_uid(uid) for e in self.aa_allele.edits])

    def get_jaccard(self, other):
        aa_jaccard = jaccard(self.aa_allele.edits, other.aa_allele.edits)
        nt_jaccard = jaccard(self.nt_allele.edits, other.nt_allele.edits)
        return (aa_jaccard, nt_jaccard)

    def get_jaccards(self, allele_list: Iterable[CodingNoncodingAllele]):
        aa_jaccards, nt_jaccards = zip(
            *list(map(lambda o: self.get_jaccard(o), allele_list))
        )
        aa_jaccards = np.array(aa_jaccards)
        nt_jaccards = np.array(nt_jaccards)
        return (aa_jaccards, nt_jaccards)

    def map_to_closest(
        self,
        allele_list,
        aa_jaccard_threshold=0.5,
        nt_jaccard_threshold=0.5,
        merge_priority: np.ndarray = None,
    ):
        """
        Arguments
        merge_priority -- Priority on which allele to merge if the jaccard index is the same.
        """
        if len(allele_list) == 0:
            return CodingNoncodingAllele()
        aa_jaccards, nt_jaccards = self.get_jaccards(allele_list)
        with warnings.catch_warnings():  # catch all NaN warning
            warnings.simplefilter("ignore")
            has_aa_jac = bool(self.aa_allele)
            has_nt_jac = bool(self.nt_allele)
            if has_aa_jac:
                aa_max_idx = np.where(aa_jaccards == np.nanmax(aa_jaccards))[0]
            else:
                aa_max_idx = np.where(
                    map(lambda cn_allele: not cn_allele.aa_allele, allele_list)
                )[0]
            if has_nt_jac:
                nt_max_idx = np.where(nt_jaccards == np.nanmax(nt_jaccards))[0]
            else:
                nt_max_idx = np.where(
                    map(lambda cn_allele: not cn_allele.nt_allele, allele_list)
                )[0]
            both_max_idx = np.intersect1d(aa_max_idx, nt_max_idx)

            def get_max_idx(max_idces):
                if len(max_idces) > 1 and not merge_priority is None:
                    if not len(merge_priority) == len(allele_list):
                        raise ValueError(
                            "merge_priority length {} is not the same as allele_list length {}".format(
                                len(merge_priority), len(allele_list)
                            )
                        )
                    return max_idces[np.argmax(merge_priority[max_idces])]
                elif len(max_idces) > 0:
                    return max_idces[0]
                else:
                    return -1

            both_max_idx = get_max_idx(both_max_idx)
            if both_max_idx >= 0:
                if (
                    has_aa_jac and aa_jaccards[both_max_idx] >= aa_jaccard_threshold
                ) or (has_nt_jac and nt_jaccards[both_max_idx] >= nt_jaccard_threshold):
                    return allele_list[both_max_idx.item()]
            elif has_aa_jac:
                aa_max_idx = get_max_idx(aa_max_idx)
                if aa_max_idx >= 0:
                    return allele_list[aa_max_idx.item()]
            elif has_nt_jac:
                nt_max_idx = get_max_idx(nt_max_idx)
                if nt_max_idx >= 0:
                    return allele_list[nt_max_idx.item()]
            return CodingNoncodingAllele()

    def __bool__(self):
        return bool(self.aa_allele) or bool(self.nt_allele)

    def __len__(self):
        return len(self.aa_allele) + len(self.nt_allele)

    def __repr__(self):
        return str(self.aa_allele) + "|" + str(self.nt_allele)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (
            self.uid == other.uid
            and self.aa_allele == other.aa_allele
            and self.nt_allele == other.nt_allele
        )

    def __lt__(self, other):
        if self.aa_allele < other.aa_allele:
            return True
        if self.aa_allele == other.aa_allele:
            if self.nt_allele < other.nt_allele:
                return True
        return False

    def __gt__(self, other):
        if self.aa_allele > other.aa_allele:
            return True
        if self.aa_allele == other.aa_allele:
            if self.nt_allele > other.nt_allele:
                return True
        return False

    def __hash__(self):
        return hash(self.__repr__())



from copy import deepcopy
from typing import Optional, Sequence
import re
from ..framework.Edit import Edit
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from tqdm.auto import tqdm

BASES = ["A", "T", "C", "G"]
REVCOMP = {"A": "T", "T": "A", "C": "G", "G": "C"}
other_from = {"A": "C", "T": "G", "C": "A", "G": "T"}


def _add_absent_edits(
    bdata,
    guide: str,
    edit_tbl: pd.DataFrame,
    edited_base: str = "A",
    target_alt: str = "G",
):
    """If edit is not observed in editable position, add into the edit rate table."""
    editable_positions = np.where(
        (np.array(list(bdata.guides.loc[guide, "Reporter"])) == edited_base)
    )[0]
    if guide not in edit_tbl.guide:
        observed_rel_pos = []
    else:
        edited_db = edit_tbl.loc[edit_tbl.guide == guide, :]
        observed_rel_pos = edited_db.rel_pos
    edits = []
    positions = []
    for editable_pos in editable_positions:
        if editable_pos not in observed_rel_pos:
            edits.append(
                Edit(
                    editable_pos,
                    edited_base,
                    target_alt,
                )
            )
            positions.append(editable_pos)
    edit_tbl = pd.concat(
        [
            edit_tbl,
            pd.DataFrame(
                {
                    "guide": guide,
                    "edit": edits,
                    "rel_pos": positions,
                    "rep_median": 0.0,
                    "rep_mean": 0.0,
                },
            ),
        ],
        ignore_index=True,
    )
    return edit_tbl


def get_edit_rates(
    bdata, edit_count_key="edit_counts", add_absent=True, adjust_spacer_pos: bool = True
):
    """
    Obtain position- and context-wise editing rate (context: base preceding the target base position).
    Args
    bdata (bean.ReporterScreen): input ReporterScreen object to be used for analyzing posiiton-wide editing rate.
    edit_count_key: key in bdata.uns with edit count DataFrame to be used for analyzing posiiton-wide editing rate.
    """
    edit_rates = bdata.get_normalized_allele_counts(bdata.uns[edit_count_key])
    edit_rates_agg = edit_rates[["guide", "edit"]]
    edit_rates_agg = edit_rates_agg.loc[
        edit_rates_agg.guide.map(lambda s: "CONTROL" not in s)
    ].reset_index(drop=True)
    try:
        edit_rates_agg["rep_median"] = edit_rates.iloc[:, 2:].median(axis=1)
        edit_rates_agg["rep_mean"] = edit_rates.iloc[:, 2:].mean(axis=1)
        edit_rates_agg["rel_pos"] = edit_rates_agg.edit.map(lambda e: e.rel_pos).astype(
            int
        )
        if add_absent:
            for guide in tqdm(
                edit_rates_agg.guide.unique(),
                desc="Calibrating edits in editable positions...",
            ):
                edit_rates_agg = _add_absent_edits(
                    bdata,
                    guide,
                    edit_rates_agg,
                    edited_base=bdata.base_edited_from,
                    target_alt=bdata.base_edited_to,
                )
        if adjust_spacer_pos:
            if "guide_len" not in bdata.guides.columns:
                bdata.guides["guide_len"] = bdata.guides.sequence.map(len)
            start_pad = (
                32
                - 6
                - bdata.guides.guide_len[edit_rates_agg.guide].reset_index(drop=True)
            )
            edit_rates_agg["spacer_pos"] = (edit_rates_agg.rel_pos - start_pad).astype(
                int
            )
            edit_rates_agg = edit_rates_agg[
                (edit_rates_agg.spacer_pos >= 0) & (edit_rates_agg.spacer_pos < 20)
            ].reset_index(drop=True)
            edit_rates_agg.loc[:, "spacer_pos"] = edit_rates_agg["spacer_pos"] + 1
        else:
            edit_rates_agg["spacer_pos"] = edit_rates_agg.rel_pos + 1
        edit_rates_agg["base_change"] = edit_rates_agg.edit.map(
            lambda e: e.get_base_change()
        )
        edit_rates_agg.rel_pos = edit_rates_agg.rel_pos.astype(int)
        edit_rates_agg["context"] = edit_rates_agg.apply(
            lambda row: bdata.guides.loc[row.guide, "Reporter"][
                row.rel_pos - 1 : row.rel_pos + 1
            ],
            axis=1,
        )
    except Exception as exp:
        print(exp)
    return edit_rates_agg


def plot_by_pos_context(
    edit_rates_df: pd.DataFrame,
    target_base="A",
):
    target_base_alt = {"A": "G", "C": "T"}[target_base]
    context_to_offset_map = {
        f"A{target_base}": -0.4,
        f"G{target_base}": -0.2,
        f"C{target_base}": 0,
        f"T{target_base}": 0.2,
    }
    edit_rates_df[
        "spacer_pos_ctxt"
    ] = edit_rates_df.spacer_pos + edit_rates_df.context.map(context_to_offset_map)
    fig, ax = plt.subplots(figsize=(6, 3))
    sns.scatterplot(
        edit_rates_df.loc[
            (edit_rates_df.base_change == f"{target_base}>{target_base_alt}")
        ],
        x="spacer_pos_ctxt",
        y="rep_mean",
        hue="context",
        alpha=0.3,
        ax=ax,
        s=5,
        rasterized=True,
    )
    ax.legend(bbox_to_anchor=(1.02, 0.5), loc="center left", title="Context")
    ax.set_xlabel("Protospacer position")
    ax.set_ylabel(f"{target_base}>{target_base_alt} editing rate")
    ax.set_ylim((0, 1))
    return ax


def _get_complementary_base_change(base_change_str: str):
    if not re.fullmatch(r"[ATCG]>[ATCG]", base_change_str):
        raise ValueError(
            f"Input argument {base_change_str} doesn't conform to the valid format ex. 'A>G'"
        )
    base_from = base_change_str[0]
    base_to = base_change_str[-1]
    return f"{REVCOMP[base_from]}>{REVCOMP[base_to]}"


def _combine_complementary_base_changes(edit_rate_df: pd.DataFrame, show_list=None):
    """Combine complementary list"""

    edit_rate_df_reduced = edit_rate_df[show_list].copy()
    for base_change in show_list:
        rev_base_change = _get_complementary_base_change(base_change)
        if rev_base_change in edit_rate_df.columns:
            edit_rate_df_reduced.loc[:, base_change] += (
                edit_rate_df[base_change] + edit_rate_df[rev_base_change]
            )
        # edit_rate_df_reduced.loc[:,base_change] = edit_rate_df_reduced.loc[:,base_change]/2
    return edit_rate_df_reduced


def _get_norm_rates_df(
    bdata,
    edit_rates_df=None,
    edit_count_key="edit_counts",
    base_changes: Sequence[str] = None,
    show_list=None,
):
    change_by_pos = pd.pivot(
        edit_rates_df.groupby(["base_change", "spacer_pos"])
        .sum()["rep_mean"]
        .reset_index(),
        index="spacer_pos",
        columns="base_change",
        values="rep_mean",
    ).fillna(0)

    norm_matrix = pd.DataFrame(index=change_by_pos.index, columns=BASES)
    for pos in norm_matrix.index:
        pos_base = bdata.guides.Reporter.map(
            lambda s: s[pos] if pos < len(s) else " "
        ).values
        for b in BASES:
            norm_matrix.loc[pos, b] = (pos_base == b).sum()
    ref_bases = change_by_pos.columns.map(lambda s: s.split(">")[0])
    change_by_pos = change_by_pos.loc[:, ref_bases.isin(BASES)]
    norm_rate = (
        change_by_pos
        / norm_matrix.loc[
            :, change_by_pos.columns.map(lambda s: s.split(">")[0])
        ].values
    )
    if show_list is None:
        show_list = ["A>C", "A>T", "A>G", "C>T", "C>G"]
    # norm_rate_reduced = _combine_complementary_base_changes(norm_rate).astype(float)
    return norm_rate.astype(float)[show_list]  # _reduced


def _get_possible_changes_from_target_base(target_basechange: str):
    """Return base changes strings (ex. A>C) for given the same reference base to edit from to the input target_basechange. ex) returns ['A>C', 'A>T'] given input 'A>G'."""
    if not re.fullmatch(r"[ATCG]>[ATCG]", target_basechange):
        raise ValueError(
            f"Input argument {target_basechange} doesn't conform to the valid format ex. 'A>G'"
        )
    base_from = target_basechange[0]
    base_to = target_basechange[-1]
    return [f"{base_from}>{b}" for b in BASES if b not in [base_from, base_to]]


def plot_by_pos_behive(
    norm_rates_df: Optional[pd.DataFrame] = None,
    bdata=None,
    edit_count_key: str = "edit_counts",
    target_basechange="A>G",
    nonref_base_changes: Sequence[str] = None,
    normalize=False,
):
    """Plot position-wise editing pattern as in BE-Hive.
    Args
    bdata (bean.ReporterScreen): input ReporterScreen object to be used for analyzing posiiton-wide editing rate.
    edit_count_key: key in bdata.uns with edit count DataFrame to be used for analyzing posiiton-wide editing rate.
    df_to_draw: input DataFrame with position-wise editing rates (n_position by n_basechange)
    norm_rates_df: pd.DataFrame with normalized editing rate with columns ['guide', 'edit', 'median/mean editing rate', 'spacer_pos']
    normalize: Normalize the editing rate relative to the max editing rate by position (as 100).

    """
    if not re.fullmatch(r"[ATCG]>[ATCG]", target_basechange):
        raise ValueError(
            f"Input argument {target_basechange} doesn't conform to the valid format ex. 'A>G'"
        )
    if nonref_base_changes is None:
        if target_basechange == "A>G":
            nonref_base_changes = ["C>T", "C>G"]
        else:
            print("No non-ref base changes specified. not drawing them")
            nonref_base_changes = []
    ref_other_changes = _get_possible_changes_from_target_base(target_basechange)

    df_to_draw = _get_norm_rates_df(
        bdata,
        norm_rates_df,
        edit_count_key,
        base_changes=[
            target_basechange,
        ]
        + ref_other_changes
        + nonref_base_changes,
    )
    fig, ax = plt.subplots(figsize=(3, 7))

    vmax = df_to_draw.max().max()
    if normalize:
        df_to_draw = df_to_draw / vmax * 100
        vmax = 100
    target_data = df_to_draw.copy()
    target_data.loc[:, target_data.columns != target_basechange] = np.nan
    sns.heatmap(
        target_data,
        ax=ax,
        annot=True,
        cmap="Reds",
        vmax=vmax,
        cbar=False,
        vmin=-0.03,
        fmt=".0f" if normalize else ".2g",
        annot_kws={"fontsize": 8},
    )
    ref_data = df_to_draw.copy()
    ref_data.loc[
        :,
        ~ref_data.columns.isin(ref_other_changes),
    ] = np.nan
    sns.heatmap(
        ref_data,
        ax=ax,
        annot=True,
        cmap="Blues",
        vmax=vmax,
        cbar=False,
        fmt=".1g",
        vmin=-0.03,
        annot_kws={"fontsize": 8},
    )

    nonref_data = df_to_draw.copy()
    nonref_data.loc[:, ~nonref_data.columns.isin(nonref_base_changes)] = np.nan
    sns.heatmap(
        nonref_data,
        ax=ax,
        annot=True,
        cmap="Greys",
        vmax=vmax,
        cbar=False,
        fmt=".1g",
        vmin=-0.03,
        annot_kws={"fontsize": 8},
    )
    ax.set_ylabel("Protospacer position")
    return ax


def get_position_by_pam_rates(bdata, edit_rates_df: pd.DataFrame):
    edit_rates_df["pam"] = bdata.guides.loc[
        edit_rates_df.guide, "5-nt PAM"
    ].reset_index(drop=True)
    edit_rates_df["pam23"] = edit_rates_df.pam.map(lambda s: s[1:3])
    edit_rates_df["pam2"] = edit_rates_df.pam.map(lambda s: s[1])
    edit_rates_df["pam3"] = edit_rates_df.pam.map(lambda s: s[2])
    edit_rates_df["pam12"] = edit_rates_df.pam.map(lambda s: s[:2])
    edit_rates_df["pam34"] = edit_rates_df.pam.map(lambda s: s[2:4])
    return pd.pivot(
        edit_rates_df.loc[(edit_rates_df.base_change == bdata.target_base_change)]
        .groupby(["pam23", "spacer_pos"])
        .mean()
        .reset_index(),
        index="pam23",
        columns="spacer_pos",
        values="rep_mean",
    )


def plot_by_pos_pam(
    bdata,
    edit_rates_df,
    ax=None,
    figsize=(6, 4),
):
    """Plot position by PAM"""
    pos_by_pam = get_position_by_pam_rates(bdata, edit_rates_df)
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    sns.heatmap(pos_by_pam, ax=ax, cmap="Blues")
    ax.set_yticklabels(ax.get_yticklabels(), rotation=0)


def plot_by_pos_pam_and_context(bdata, edit_rates_df, figsize=(6, 6)):
    pos_by_pam = get_position_by_pam_rates(bdata, edit_rates_df)
    fig, ax = plt.subplots(2, 1, figsize=figsize)
    sns.scatterplot(
        edit_rates_df.loc[(edit_rates_df.base_change == bdata.target_base_change)],
        x="spacer_pos_ctxt",
        y="rep_mean",
        hue="context",
        alpha=0.3,
        ax=ax[0],
        s=5,
        rasterized=True,
    )
    ax[0].legend(bbox_to_anchor=(1.02, 0.5), loc="center left", title="Context")
    ax[0].set_xlabel("Protospacer position")
    ax[0].set_xticks(list(range(1, 21)))
    ax[0].set_xticklabels(list(range(1, 21)))
    ax[0].set_ylabel(f"{bdata.target_base_change} editing rate")
    ax[0].set_ylim((0, 1))
    ax[0].set_xlim((0.5, 20.5))

    sns.heatmap(pos_by_pam, ax=ax[1], cmap="Blues", cbar=False)
    ax[1].set_yticklabels(ax[1].get_yticklabels(), rotation=0)
    cbax = fig.add_axes(
        [0.85, 0.15, 0.02, 0.3],
    )
    ax[1].set_ylabel("PAM")
    ax[1].set_yticklabels([f"N{t.get_text()}" for t in ax[1].get_yticklabels()])
    ax[1].set_xlabel("Protospacer position")
    fig.colorbar(ax[1].get_children()[0], cbax, label="editing rate")
    plt.tight_layout()
    return ax


def get_pam_preference(
    edit_rates_df: pd.DataFrame,
    edit_start_pos: int = 3,
    edit_end_pos: int = 8,
    norm=True,
):
    window_edits = deepcopy(
        edit_rates_df.loc[
            (edit_rates_df.spacer_pos >= edit_start_pos)
            & (edit_rates_df.spacer_pos < edit_end_pos)
            & (edit_rates_df.base_change == "A>G")
        ]
    )
    if norm:
        ctxt_mean = window_edits.groupby("context")["rep_mean"].mean()
        ctxt_mean = ctxt_mean / ctxt_mean.max()
        pos_mean = window_edits.groupby("spacer_pos")["rep_mean"].mean()
        pos_mean = pos_mean / pos_mean.max()
        print("Normalizing per context mean")
        window_edits["norm_rate"] = (
            window_edits["rep_mean"]
            / ctxt_mean.loc[window_edits.context].values
            / pos_mean.loc[window_edits.spacer_pos].values
        )
    else:
        window_edits["norm_rate"] = window_edits.rep_mean
    pam_pref = pd.pivot(
        window_edits.groupby(["pam2", "pam3"]).mean().reset_index(),
        columns="pam3",
        index="pam2",
        values="norm_rate",
    )
    return pam_pref


def plot_pam_preference(
    edit_rates_df,
    bdata=None,
    edit_start_pos: int = 3,
    edit_end_pos: int = 8,
    ax=None,
    norm=True,
):
    """ """
    if "pam2" not in edit_rates_df.columns or "pam3" not in edit_rates_df.columns:
        edit_rates_df["pam"] = bdata.guides.loc[
            edit_rates_df.guide, "5-nt PAM"
        ].reset_index(drop=True)
        edit_rates_df["pam2"] = edit_rates_df.pam.map(lambda s: s[1])
        edit_rates_df["pam3"] = edit_rates_df.pam.map(lambda s: s[2])
    pam_pref = get_pam_preference(
        edit_rates_df, edit_start_pos, edit_end_pos, norm=norm
    )
    fig, ax = plt.subplots(figsize=(3, 3))
    sns.heatmap(data=pam_pref, annot=True, cmap="Blues")
    ax.set_box_aspect(1)
    ax.set_xticklabels([f"NN{t.get_text()}" for t in ax.get_xticklabels()])
    ax.set_yticklabels([f"N{t.get_text()}N" for t in ax.get_yticklabels()])
    ax.set(xlabel=None, ylabel=None)
    ax.set_title("PAM preference")
    return ax



import os
from typing import List, Iterable, Dict, Tuple, Collection
from copy import deepcopy
import numpy as np
import pandas as pd
from Bio import SeqIO
import bean as be
from bean.framework.Edit import Edit, Allele
from bean.framework.AminoAcidEdit import (
    AminoAcidEdit,
    CodingNoncodingAllele,
)
import logging
import sys

logging.basicConfig(
    level=logging.INFO,
    format="%(levelname)-5s @ %(asctime)s:\n\t %(message)s \n",
    datefmt="%a, %d %b %Y %H:%M:%S",
    stream=sys.stderr,
    filemode="w",
)
error = logging.critical
warn = logging.warning
debug = logging.debug
info = logging.info

BASE_SET = {"A", "C", "T", "G"}
reverse_map = {"A": "T", "C": "G", "G": "C", "T": "A", "N": "N", "-": "-"}

codon_map = {
    "TTT": "F",
    "TTC": "F",
    "TTA": "L",
    "TTG": "L",
    "CTT": "L",
    "CTC": "L",
    "CTA": "L",
    "CTG": "L",
    "ATT": "I",
    "ATC": "I",
    "ATA": "I",
    "ATG": "M",
    "GTT": "V",
    "GTC": "V",
    "GTA": "V",
    "GTG": "V",
    "TCT": "S",
    "TCC": "S",
    "TCA": "S",
    "TCG": "S",
    "CCT": "P",
    "CCC": "P",
    "CCA": "P",
    "CCG": "P",
    "ACT": "T",
    "ACC": "T",
    "ACA": "T",
    "ACG": "T",
    "GCT": "A",
    "GCC": "A",
    "GCA": "A",
    "GCG": "A",
    "TAT": "Y",
    "TAC": "Y",
    "TAA": "*",
    "TAG": "*",
    "CAT": "H",
    "CAC": "H",
    "CAA": "Q",
    "CAG": "Q",
    "AAT": "N",
    "AAC": "N",
    "AAA": "K",
    "AAG": "K",
    "GAT": "D",
    "GAC": "D",
    "GAA": "E",
    "GAG": "E",
    "TGT": "C",
    "TGC": "C",
    "TGA": "*",
    "TGG": "W",
    "CGT": "R",
    "CGC": "R",
    "CGA": "R",
    "CGG": "R",
    "AGT": "S",
    "AGC": "S",
    "AGA": "R",
    "AGG": "R",
    "GGT": "G",
    "GGC": "G",
    "GGA": "G",
    "GGG": "G",
}


class RefBaseMismatchException(Exception):
    pass


# This function is adopted from https://github.com/gpp-rnd/be-validation-pipeline/blob/main/notebooks/01_BEV_allele_frequencies.ipynb
def _translate(seq: str, codon_map: Dict[str, str]):
    """Translate seq"""
    #     if not seq: # if remove_introns returned False -> possible splice site mutation
    if seq == "UTR":
        return "Possible UTR mutation"
    if seq == "intron":
        return "Possible intronic mutation"
    aa = ""
    i = 0
    frame = 1
    while i < len(seq):
        substring = ""
        while frame <= 3:
            if i >= len(seq):
                return aa + ">"
            if seq[i] != "-":
                substring += seq[i]
                frame += 1
            i += 1
        if len(substring) == 3:
            aa = aa + "-" if "N" in substring else aa + codon_map[substring]
        frame = 1  # reset frame
    return aa


def _parse_range(chr_range: str) -> Tuple[str, int, int]:
    """Parse 'chr:start-end' into chr,start,end tuple"""
    chrom, gen_range = chr_range.split(":")
    start, end = gen_range.split("-")
    start = int(start)
    end = int(end)
    return (chrom, start, end)


def _parse_description(desc_str: str):
    """Parse description line of fasta file to get range."""
    sstr = desc_str.split(" ")
    for s in sstr:
        if s.startswith("range="):
            return _parse_range(s[6:])


def _get_seq_pos_from_fasta(fasta_file_name: str) -> Tuple[List[str], List[int]]:
    """Obtain tuple of lists, first with nucleotide and second with genomic position of the nucleotide."""
    exons = list(SeqIO.parse(fasta_file_name, "fasta"))
    translated_seq = []
    genomic_pos = []
    for exon in exons:
        _, exon_start, exon_end = _parse_description(exon.description)
        for i, nt in enumerate(str(exon.seq)):
            if nt.islower():
                continue
            translated_seq.append(nt)
            genomic_pos.append(exon_start + i)
    return (translated_seq, genomic_pos)


def _translate_single_codon(nt_seq_string: str, aa_pos: int) -> str:
    """Translate `aa_pos`-th codon of `nt_seq_string`."""
    codon = "".join(nt_seq_string[aa_pos * 3 : (aa_pos * 3 + 3)])
    if len(codon) != 3:
        print("reached the end of CDS, frameshift.")
        return "/"
    try:
        return codon_map[codon]
    except KeyError:
        if codon[-1] == "N" and codon[0] in BASE_SET and codon[1] in BASE_SET:
            aa_set = {codon_map[codon[:2] + N] for N in BASE_SET}
            if len(aa_set) == 1:
                return next(iter(aa_set))
            print(f"warning: no matching aa with codon {codon}")
        else:
            print(f"Cannot translate codon due to ambiguity: {codon}")

        return "_"


class CDS:
    def __init__(self, fasta_file_name=None, suppressMessage=True):
        if fasta_file_name is None:
            if not suppressMessage:
                print("No fasta file provided as reference: using LDLR")
            fasta_file_name = os.path.dirname(be.__file__) + "/annotate/ldlr_exons.fa"
        try:
            type(self).set_exon_fasta_name(fasta_file_name)
        except FileNotFoundError as e:
            print(os.getcwd())
            raise e
        self.edited_nt = type(self).nt.copy()
        self.edited_aa_pos = set()
        self.edits_noncoding = set()
        self.set_exon_fasta_name(fasta_file_name)

    @classmethod
    def set_exon_fasta_name(cls, fasta_file_name: str):
        cls.fasta_file_name = fasta_file_name
        cls.translated_seq, cls.genomic_pos = _get_seq_pos_from_fasta(fasta_file_name)
        cls.nt = cls.translated_seq
        cls.pos = np.array(cls.genomic_pos)

    def translate(self):
        self.aa = _translate(self.edited_nt, codon_map)

    def _get_relative_nt_pos(self, absolute_pos):
        nt_relative_pos = np.where(type(self).pos == absolute_pos)[0]
        assert len(nt_relative_pos) <= 1, nt_relative_pos
        return nt_relative_pos.astype(int).item() if nt_relative_pos else -1

    def _edit_pos_to_aa_pos(self, edit_pos):
        nt_relative_pos = self._get_relative_nt_pos(edit_pos)
        if nt_relative_pos != -1:
            self.edited_aa_pos.add(nt_relative_pos // 3)
        return nt_relative_pos

    def edit_single(self, edit_str):
        edit = Edit.from_str(edit_str)
        rel_pos = self._edit_pos_to_aa_pos(edit.pos)
        if edit.strand == "-":
            ref_base = reverse_map[edit.ref_base]
            alt_base = reverse_map[edit.alt_base]
        else:
            ref_base = edit.ref_base
            alt_base = edit.alt_base
        if rel_pos == -1:  # position outside CDS
            self.edits_noncoding.add(edit)
        elif type(self).nt[rel_pos] != ref_base:
            if ref_base != "-":
                raise RefBaseMismatchException(
                    f"ref:{type(self).nt[rel_pos]} at pos {rel_pos}, got edit {edit}"
                )
        else:
            self.edited_nt[rel_pos] = alt_base
        if alt_base == "-":  # frameshift
            # self.edited_nt.pop(rel_pos)
            self.edited_aa_pos.update(list(range(rel_pos, len(CDS.nt) // 3)))

    def edit_allele(self, allele_str):
        if isinstance(allele_str, Allele):
            edit_strs = allele_str.edits
        else:
            edit_strs = allele_str.split(",")
        for edit_str in edit_strs:
            self.edit_single(edit_str)
        if "-" in self.edited_nt:
            self.edited_nt.remove("-")

    def get_aa_change(self, include_synonymous=True) -> List[str]:
        """1-based amino acid editing result"""
        mutations = CodingNoncodingAllele()
        mutations.nt_allele.update(self.edits_noncoding)
        for edited_aa_pos in self.edited_aa_pos:
            ref_aa = _translate_single_codon(type(self).nt, edited_aa_pos)
            mt_aa = _translate_single_codon(self.edited_nt, edited_aa_pos)
            if mt_aa == "_":
                return "translation error"
            if not include_synonymous and ref_aa == mt_aa:
                continue
            mutations.aa_allele.add(AminoAcidEdit(edited_aa_pos + 1, ref_aa, mt_aa))

        return mutations


def get_allele_aa_change(allele_str, include_synonymous=True, fasta_file=None):
    """
    Obtain amino acid changes
    """
    cds = CDS(fasta_file_name=fasta_file)
    cds.edit_allele(allele_str)
    return cds.get_aa_change(include_synonymous)


def translate_allele(
    allele: Allele, include_synonymouns=True, allow_ref_mismatch=True, fasta_file=None
):
    try:
        return get_allele_aa_change(
            allele, include_synonymous=include_synonymouns, fasta_file=fasta_file
        )
    except RefBaseMismatchException as e:
        if not allow_ref_mismatch:
            raise e
        print(e)
        return "ref mismatch"


def translate_allele_df(
    allele_df, include_synonymouns=True, allow_ref_mismatch=True, fasta_file=None
):
    allele_df = allele_df.copy()
    translated_alleles = allele_df.allele.map(
        lambda a: be.translate_allele(
            a,
            include_synonymouns=include_synonymouns,
            allow_ref_mismatch=allow_ref_mismatch,
            fasta_file=fasta_file,
        )
    )
    allele_df.insert(2, "cn_allele", translated_alleles)
    allele_df["cn_allele"] = allele_df.cn_allele.map(str)
    allele_df = allele_df.loc[
        ~allele_df.cn_allele.map(str).isin(
            ["ref mismatch", "|", "", "translation error"]
        ),
        :,
    ]
    allele_df = allele_df.drop("allele", axis=1).groupby(["guide", "cn_allele"]).sum()
    allele_df = allele_df.reset_index().rename(columns={"cn_allele": "aa_allele"})
    allele_df.aa_allele = allele_df.aa_allele.map(
        lambda s: be.CodingNoncodingAllele.from_str(s)
    )
    return allele_df


def filter_nt_allele(cn_allele: CodingNoncodingAllele, pos_include: Iterable[int]):
    """
    For CodingNoncodingAllele object, retain all amino acid mutation while filtering the nt_allele based on edit position.
    """
    cn_allele = deepcopy(cn_allele)
    edit_list = [e for e in cn_allele.nt_allele.edits if e.pos in pos_include]
    cn_allele.nt_allele = be.Allele(edit_list)
    return cn_allele


def filter_nt_alleles(cn_allele_df: pd.DataFrame, pos_include: Iterable[int]):
    """
    For CodingNoncodingAllele object, retain all amino acid mutation while filtering the nt_allele based on edit position.
    Arguments
    -- cn_allele_df (pd.DataFrame): Allele dataframe that has 'guide', 'aa_allele' as columns.
    """
    splice_only = cn_allele_df.aa_allele.map(
        lambda a: be.an.translate_allele.filter_nt_allele(a, pos_include)
    )
    alleles = cn_allele_df.copy()
    alleles = alleles.drop("aa_allele", axis=1, inplace=False)
    alleles.insert(1, "aa_allele", splice_only)
    alleles = alleles.groupby(["guide", "aa_allele"]).sum().reset_index()
    alleles = alleles.loc[alleles.aa_allele.map(bool), :]
    return alleles


def annotate_edit(
    edit_info: pd.DataFrame,
    edit_col="edit",
    splice_sites: Collection[
        int
    ] = None,  # TODO: may be needed to extended into multi-chromosome case
):
    """Classify edit strings into coding/noncoding and synonymous/misseinse[splicing/trunc]

    Args
    edit_info: pd.DataFrame with at least 1 column of 'edit_col', which has 'Edit' format.
    splice_sites: Collection of integer splice site positions. If the edit position matches the positions, it will be annotated as 'splicing'.

    """
    edit_info = edit_info.copy()
    edit_info["group"] = ""
    edit_info["int_pos"] = -1
    if "pos" not in edit_info.columns:
        edit_info["pos"], transition = zip(*(edit_info[edit_col].str.split(":")))
        edit_info["ref"], edit_info["alt"] = zip(
            *(pd.Series(transition).str.split(">"))
        )
    edit_info.loc[edit_info.pos.map(lambda s: s.startswith("A")), "coding"] = "coding"
    edit_info.loc[
        edit_info.pos.map(lambda s: not s.startswith("A")), "coding"
    ] = "noncoding"
    edit_info.loc[edit_info.pos.map(lambda s: "CONTROL" in s), "group"] = "negctrl"
    edit_info.loc[edit_info.pos.map(lambda s: "CONTROL" in s), "coding"] = "negctrl"
    edit_info.loc[
        (edit_info.coding == "noncoding") & (edit_info.group != "negctrl"), "int_pos"
    ] = edit_info.loc[
        (edit_info.coding == "noncoding") & (edit_info.group != "negctrl"), "pos"
    ].map(
        int
    )

    edit_info.loc[
        (edit_info.alt != edit_info.ref) & (edit_info.coding == "coding"), "group"
    ] = "missense"
    edit_info.loc[edit_info.alt == "*", "group"] = "trunc"
    edit_info.loc[
        (edit_info.alt == edit_info.ref) & (edit_info.coding == "coding"), "group"
    ] = "syn"
    if splice_sites is not None:
        edit_info.loc[
            edit_info.pos.isin(splice_sites.astype(str)), "group"
        ] = "splicing"
    edit_info.loc[edit_info.int_pos < -100, "group"] = "negctrl"
    edit_info.loc[edit_info.int_pos < -100, "coding"] = "negctrl"
    return edit_info



import numpy as np
from setuptools import setup, Extension, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="crispr-bean",
    version="0.2.2",
    python_requires=">=3.8.0",
    author="Jayoung Ryu",
    author_email="jayoung_ryu@g.harvard.edu",
    description="Base Editor screen analysis [Bayesian Estimation of variant effect] with guide Activity Normalization",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pinellolab/crispr-bean",
    packages=find_packages(),
    ext_modules=[Extension("bean", sources=["bean/mapping/CRISPResso2Align.pyx"])],
    include_dirs=np.get_include(),
    setup_requires=[
        "setuptools>=18.0",
        "cython",
    ],
    scripts=[
        "bin/bean-count",
        "bin/bean-count-samples",
        "bin/bean-qc",
        "bin/bean-filter",
        "bin/bean-run",  # TODO: prevent error when extra requirements are not met.
    ],
    install_requires=[
        "numpy",
        "perturb-tools>=0.2.0",
        "anndata>=0.8.0",
        "Bio>=1.5",
        "matplotlib",
        "pandas",
        "scipy",
        "seaborn",
        "tqdm",
    ],
    extras_require={"model": ["pyBigWig", "pyro-ppl<=1.8.1", "statsmodels", "torch<2"]},
    include_package_data=True,
    package_data={
        "": [
            "bean/annotate/ldlr_exons.fa",
            "notebooks/sample_quality_report.ipynb",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)
